@@ -2,7 +2,7 @@
 
 ;;; Commentary:
 
-;; 
+;;
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;; Dependency
@@ -10,7 +10,7 @@
 
 ;;; Code:
 
-(defvar agda2-version "2.2.4"
+(defvar agda2-version "2.2.6"
   "The version of the Agda mode.
 Note that, by default, the same version of the underlying Haskell
 library is used (see `agda2-ghci-options').")
@@ -81,7 +81,8 @@ All the elements of XS are included, in their original order."
 (defcustom agda2-include-dirs
   '(".")
   "The directories Agda uses to search for files.
-The directory names should be relative to the root of the current project."
+The directory names should either be absolute or be relative to
+the root of the current project."
   :type '(repeat directory)
   :group 'agda2)
 
@@ -226,11 +227,10 @@ constituents.")
     (agda2-load                              "\C-c\C-l"         (global)       "Load")
     (agda2-load                              "\C-c\C-x\C-l")
     (agda2-compile                           "\C-c\C-x\C-c"     (global)       "Compile")
-    (agda2-text-state                        "\C-c\C-x\C-d"     (global)       "Deactivate Agda")
     (agda2-quit                              "\C-c\C-x\C-q"     (global)       "Quit")
     (agda2-restart                           "\C-c\C-x\C-r"     (global)       "Restart")
+    (agda2-remove-annotations                "\C-c\C-x\C-d"     (global)       "Remove goals and highlighting (\"deactivate\")")
     (agda2-display-implicit-arguments        "\C-c\C-x\C-h"     (global)       "Toggle display of hidden arguments")
-    (agda2-highlight-reload-or-clear         "\C-c\C-x\C-s"     (global)       "Reload syntax highlighting information")
     (agda2-show-constraints                  ,(kbd "C-c C-=")   (global)       "Show constraints")
     (agda2-solveAll                          ,(kbd "C-c C-s")   (global)       "Solve constraints")
     (agda2-show-goals                        ,(kbd "C-c C-?")   (global)       "Show goals")
@@ -238,6 +238,7 @@ constituents.")
     (agda2-previous-goal                     "\C-c\C-b"         (global)       "Previous goal") ; Back.
     (agda2-give                              ,(kbd "C-c C-SPC") (local)        "Give")
     (agda2-refine                            "\C-c\C-r"         (local)        "Refine")
+    (agda2-auto                              "\C-c\C-a"         (local)        "Auto")
     (agda2-make-case                         "\C-c\C-c"         (local)        "Case")
     (agda2-goal-type                         "\C-c\C-t"         (local)        "Goal type")
     (agda2-show-context                      "\C-c\C-e"         (local)        "Context (environment)")
@@ -328,15 +329,17 @@ menus.")
 (modify-coding-system-alist 'file "\\.l?agda\\'" 'utf-8)
 ;;;###autoload
 (define-derived-mode agda2-mode nil "Agda"
- "Major mode for agda files.
+ "Major mode for Agda files.
 
-Note that when this mode is activated the default font of the
-current frame is changed to the fontset `agda2-fontset-name'.
-The reason is that Agda programs often use mathematical symbols
-and other Unicode characters, so we try to provide a suitable
-default font setting, which can display many of the characters
-encountered.  If you prefer to use your own settings, set
-`agda2-fontset-name' to nil.
+The following paragraph does not apply to Emacs 23.
+
+  Note that when this mode is activated the default font of the
+  current frame is changed to the fontset `agda2-fontset-name'.
+  The reason is that Agda programs often use mathematical symbols
+  and other Unicode characters, so we try to provide a suitable
+  default font setting, which can display many of the characters
+  encountered. If you prefer to use your own settings, set
+  `agda2-fontset-name' to nil.
 
 Special commands:
 \\{agda2-mode-map}"
@@ -353,13 +356,23 @@ Special commands:
          (set-frame-font agda2-fontset-name)
        (error (error "Unable to change the font; change agda2-fontset-name or tweak agda2-fontset-spec-of-fontset-agda2"))))
  (agda2-indent-setup)
+ ;; If GHCi is not running syntax highlighting does not work properly.
+ (unless (eq 'run (agda2-process-status))
+   (agda2-restart))
  (agda2-highlight-setup)
  (agda2-highlight-reload)
  (agda2-comments-and-paragraphs-setup)
  (force-mode-line-update)
  ;; Protect global value of default-input-method from set-input-method.
  (make-local-variable 'default-input-method)
- (set-input-method "Agda"))
+ (set-input-method "Agda")
+ ;; Highlighting is removed when we switch from the Agda mode. Use
+ ;; case: When a file M.lagda with a local variables list including
+ ;; "mode: latex" is loaded chances are that the Agda mode is
+ ;; activated before the LaTeX mode, and the LaTeX mode does not seem
+ ;; to remove the text properties set by the Agda mode.
+ (add-hook (make-local-variable 'change-major-mode-hook)
+           'agda2-remove-annotations))
 
 (defun agda2-restart ()
   "Kill and restart the *ghci* buffer and load `agda2-toplevel-module'."
@@ -383,10 +396,11 @@ Special commands:
                          'agda2-send)
                     (set-buffer-file-coding-system 'utf-8)
                     (set-buffer-process-coding-system 'utf-8 'utf-8)
-                    (rename-buffer agda2-bufname)))
-  (apply 'agda2-go nil ":set" agda2-ghci-options)
-  (agda2-go nil ":mod +" agda2-toplevel-module)
-  (agda2-text-state))
+                    (rename-buffer agda2-bufname)
+                    (set-process-query-on-exit-flag agda2-process nil)))
+  (apply 'agda2-call-ghci ":set" agda2-ghci-options)
+  (agda2-call-ghci ":mod +" agda2-toplevel-module)
+  (agda2-remove-annotations))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;; Communicating with Agda
@@ -406,12 +420,11 @@ long strings (some versions of GHCi, on some systems)."
       (comint-send-string proc chunk)))
   (comint-send-string proc "\n"))
 
-(defun agda2-go (require-response &rest args)
+(defun agda2-call-ghci (&rest args)
   "Executes commands in GHCi.
 Sends the list of strings ARGS to GHCi, waits for output and
-executes the responses, if any. If REQUIRE-RESPONSE is non-nil
-then an error is raised if no responses are received."
-  (interactive)
+executes the responses, if any. Returns the number of processed
+responses."
   (unless (eq 'run (agda2-process-status))
     ;; Try restarting automatically, but only once, in case there is
     ;; some major problem.
@@ -419,7 +432,14 @@ then an error is raised if no responses are received."
     (unless (eq 'run (agda2-process-status))
       (agda2-raise-ghci-error)))
   (save-excursion
-    (haskell-ghci-go (apply 'concat (agda2-intersperse " " args)) nil))
+    ;; By setting local-enable-local-variables to nil the call to
+    ;; hack-local-variables in haskell-ghci-go becomes more or less a
+    ;; no-op. (Note that hack-local-variables can interfere with the
+    ;; setup of a mode, because it can potentially perform the setup
+    ;; of another mode...)
+    (let ((local-enable-local-variables nil))
+      (haskell-ghci-go (apply 'concat (agda2-intersperse " " args))
+                       nil)))
   (let (response)
     (with-current-buffer haskell-ghci-process-buffer
       (haskell-ghci-wait-for-output)
@@ -433,23 +453,68 @@ then an error is raised if no responses are received."
                 (setq response (buffer-substring-no-properties
                                 (point-min) (point-max)))))
           (delete-file tempfile))))
-    (when (and (>= 0 (agda2-respond response))
-               require-response)
-      (agda2-raise-ghci-error))))
+    (agda2-respond response)))
 
-(defun agda2-goal-cmd (require-response cmd &optional want ask &rest args)
-  "When in a goal, send CMD, goal num and range, and strings ARGS to agda2.
-WANT is an optional prompt. When ASK is non-nil, use minibuffer.
-If REQUIRE-RESPONSE is non-nil then an error is raised if no
-responses are received."
+(defun agda2-go (highlight require-response update-goals &rest args)
+  "Executes commands in GHCi.
+Sends the list of strings ARGS to GHCi, waits for output and
+executes the responses, if any. An error is raised if no
+responses are received (unless REQUIRE-RESPONSE is nil), and
+otherwise the syntax highlighting information is reloaded (unless
+HIGHLIGHT is nil; if HIGHLIGHT is a string, then highlighting
+info is read from the corresponding file) and the goals
+updated (unless UPDATE-GOALS is nil)."
+  (let* ((highlighting-temp (and highlight (not (stringp highlight))))
+         (highlighting (cond ((stringp highlight) highlight)
+                             (highlighting-temp (make-temp-file "agda2-mode")))))
+        (unwind-protect
+            (let ((no-responses
+                   (apply 'agda2-call-ghci
+                          "ioTCM"
+                          (agda2-string-quote (buffer-file-name))
+                          (if highlighting-temp
+                              (concat "(Just "
+                                      (agda2-string-quote highlighting)
+                                      ")")
+                            "Nothing")
+                          "("
+                          (append args '(")")))))
+                (when (and require-response (>= 0 no-responses))
+                  (agda2-raise-ghci-error))
+                (if highlight (agda2-highlight-load highlighting)))
+          (if highlighting-temp (delete-file highlighting))))
+  (if update-goals (agda2-annotate)))
+
+(defun agda2-goal-cmd (cmd &optional want ask &rest args)
+  "Reads input from goal or minibuffer and sends command to Agda.
+
+An error is raised if point is not in a goal.
+
+The command sent to Agda is
+
+  CMD <goal number> <goal range> <user input> ARGS.
+
+The user input is computed as follows:
+
+* If WANT is nil, then the user input is the empty string.
+
+* If WANT is a string, and either ASK is non-nil or the goal only
+  contains whitespace, then the input is taken from the
+  minibuffer. In this case WANT is used as the prompt string.
+
+* Otherwise (including if WANT is 'goal) the goal contents is
+  used.
+
+An error is raised if no responses are received."
   (multiple-value-bind (o g) (agda2-goal-at (point))
     (unless g (error "For this command, please place the cursor in a goal"))
     (let ((txt (buffer-substring-no-properties (+ (overlay-start o) 2)
                                                (- (overlay-end   o) 2))))
-      (if (not want) (setq txt "")
-          (when (or ask (string-match "\\`\\s *\\'" txt))
-            (setq txt (read-string (concat want ": ") txt))))
-      (apply 'agda2-go require-response cmd
+      (cond ((null want) (setq txt ""))
+            ((and (stringp want)
+                  (or ask (string-match "\\`\\s *\\'" txt)))
+             (setq txt (read-string (concat want ": ") txt))))
+      (apply 'agda2-go nil t t cmd
              (format "%d" g)
              (agda2-goal-Range o)
              (agda2-string-quote txt) args))))
@@ -472,13 +537,25 @@ sexp is executed. The number of executed responses is returned."
         (eval (read response))))
     no-responses))
 
+(defvar agda2-response nil
+  "Used by the backend to give responses to the Agda mode.")
+(make-variable-buffer-local 'agda2-response)
+
+(defun agda2-ask (&rest args)
+  "Executes a query in GHCi and returns the response string."
+  (setq agda2-response 'agda2-no-response)
+  (apply 'agda2-go nil nil nil args)
+  (if (equal agda2-response 'agda2-no-response)
+      (agda2-raise-ghci-error)
+    agda2-response))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;; User commands and response processing
 
 (defun agda2-load ()
   "Load current buffer."
   (interactive)
-  (agda2-go t "cmd_load"
+  (agda2-go t t t "cmd_load"
             (agda2-string-quote (buffer-file-name))
             (agda2-list-quote agda2-include-dirs)
             ))
@@ -486,32 +563,41 @@ sexp is executed. The number of executed responses is returned."
 (defun agda2-compile ()
   "Compile the current module."
   (interactive)
-  (agda2-go t "cmd_compile"
+  (agda2-go t t t "cmd_compile"
             (agda2-string-quote (buffer-file-name))
             (agda2-list-quote agda2-include-dirs)
             ))
 
-(defun agda2-load-action (gs)
-  "Annotate new goals GS in current buffer."
-  (agda2-annotate gs))
-
 (defun agda2-give()
   "Give to the goal at point the expression in it" (interactive)
-  (agda2-goal-cmd t "cmd_give" "expression to give"))
+  (agda2-goal-cmd "cmd_give" "expression to give"))
 
 (defun agda2-give-action (old-g paren)
   "Update the goal OLD-G with the expression in it."
   (agda2-update old-g paren))
 
 (defun agda2-refine ()
-  "Refine the goal at point by the expression in it." (interactive)
-  (agda2-goal-cmd t "cmd_refine" "expression to refine"))
+  "Refine the goal at point.
+If the goal contains an expression e, and some \"suffix\" of the
+type of e unifies with the goal type, then the goal is replaced
+by e applied to a suitable number of new goals.
+
+If the goal is empty, the goal type is a data type, and there is
+exactly one constructor which unifies with this type, then the
+goal is replaced by the constructor applied to a suitable number
+of new goals."
+  (interactive)
+  (agda2-goal-cmd "cmd_refine_or_intro" 'goal))
+
+(defun agda2-auto ()
+ "Simple proof search" (interactive)
+ (agda2-goal-cmd "cmd_auto"))
 
 (defun agda2-make-case ()
   "Refine the pattern var given in the goal.
 Assumes that <clause> = {!<var>!} is on one line."
   (interactive)
-  (agda2-goal-cmd t "cmd_make_case" "partten var to case"))
+  (agda2-goal-cmd "cmd_make_case" "pattern var to case"))
 
 (defun agda2-make-case-action (newcls)
   "Replace the line at point with new clauses NEWCLS and reload."
@@ -556,17 +642,17 @@ in the buffer's mode line."
 
 (defun agda2-show-goals()
   "Show all goals." (interactive)
-  (agda2-go t "cmd_metas"))
+  (agda2-go t t t "cmd_metas"))
 
 (defun agda2-show-constraints()
   "Show constraints." (interactive)
-  (agda2-go t "cmd_constraints"))
+  (agda2-go t t t "cmd_constraints"))
 
-(defun agda2-text-state ()
-  "UNDER CONSTRUCTION" (interactive)
+(defun agda2-remove-annotations ()
+  "Removes buffer annotations (overlays and text properties)."
+  (interactive)
   (dolist (o (overlays-in (point-min) (point-max)))
     (delete-overlay o))
-  (agda2-go nil "cmd_reset")
   (let ((inhibit-read-only t))
     (annotation-preserve-mod-p-and-undo
      (set-text-properties (point-min) (point-max) '()))
@@ -584,17 +670,18 @@ in the buffer's mode line."
     (or (go (point)) (go wrapped) (message "No goals in the buffer"))))
 
 (defun agda2-quit ()
-  "Quit and clean up after agda2." (interactive)
+  "Quit and clean up after agda2."
+  (interactive)
   (agda2-protect (progn (kill-buffer agda2-buffer)
                         (kill-buffer (current-buffer)))))
- 
-(defmacro agda2-maybe-normalised (name comment cmd prompt)
+
+(defmacro agda2-maybe-normalised (name comment cmd want)
   "This macro constructs a function NAME which runs CMD.
 COMMENT is used to build the function's comment. The function
 NAME takes a prefix argument which tells whether it should
 normalise types or not when running CMD (through
-`agda2-goal-cmd' t; PROMPT, if non-nil, is used as the goal
-command prompt)."
+`agda2-goal-cmd'; WANT is used as `agda2-goal-cmd's WANT
+argument)."
   (let ((eval (make-symbol "eval")))
   `(defun ,name (&optional not-normalise)
      ,(concat comment ".
@@ -602,15 +689,16 @@ command prompt)."
 With a prefix argument the result is not explicitly normalised.")
      (interactive "P")
      (let ((,eval (if not-normalise "Instantiated" "Normalised")))
-       (agda2-goal-cmd t (concat ,cmd " Agda.Interaction.BasicOps." ,eval)
-                         ,prompt)))))
+       (agda2-goal-cmd (concat ,cmd " Agda.Interaction.BasicOps." ,eval)
+                       ,want)))))
 
 (defmacro agda2-maybe-normalised-toplevel (name comment cmd prompt)
   "This macro constructs a function NAME which runs CMD.
 COMMENT is used to build the function's comments. The function
 NAME takes a prefix argument which tells whether it should
-normalise types or not when running CMD (through `agda2-go' t; the
-string PROMPT is used as the goal command prompt)."
+normalise types or not when running CMD (through
+`agda2-go' nil t nil; the string PROMPT is used as the goal
+command prompt)."
   (let ((eval (make-symbol "eval")))
     `(defun ,name (not-normalise expr)
        ,(concat comment ".
@@ -618,8 +706,9 @@ string PROMPT is used as the goal command prompt)."
 With a prefix argument the result is not explicitly normalised.")
        (interactive ,(concat "P\nM" prompt ": "))
        (let ((,eval (if not-normalise "Instantiated" "Normalised")))
-         (agda2-go t (concat ,cmd " Agda.Interaction.BasicOps." ,eval " "
-                             (agda2-string-quote expr)))))))
+         (agda2-go nil t nil
+                   (concat ,cmd " Agda.Interaction.BasicOps." ,eval " "
+                           (agda2-string-quote expr)))))))
 
 (agda2-maybe-normalised
  agda2-goal-type
@@ -671,7 +760,7 @@ top-level scope."
 (defun agda2-solveAll ()
   "Solves all goals that are already instantiated internally."
   (interactive)
-  (agda2-go t "cmd_solveAll"))
+  (agda2-go t t t "cmd_solveAll"))
 
 (defun agda2-solveAll-action (iss)
   (save-excursion
@@ -687,7 +776,7 @@ With a prefix argument \"abstract\" is ignored during the computation."
   (interactive "P")
   (let ((cmd (concat "cmd_compute"
                      (if arg " True" " False"))))
-    (agda2-goal-cmd t cmd "expression to normalise")))
+    (agda2-goal-cmd cmd "expression to normalise")))
 
 (defun agda2-compute-normalised-toplevel (expr &optional arg)
   "Computes the normal form of the given expression.
@@ -698,7 +787,7 @@ With a prefix argument \"abstract\" is ignored during the computation."
   (let ((cmd (concat "cmd_compute_toplevel"
                      (if arg " True" " False")
                      " ")))
-    (agda2-go t (concat cmd (agda2-string-quote expr)))))
+    (agda2-go nil t nil (concat cmd (agda2-string-quote expr)))))
 
 (defun agda2-compute-normalised-maybe-toplevel ()
   "Computes the normal form of the given expression,
@@ -713,12 +802,29 @@ With a prefix argument \"abstract\" is ignored during the computation."
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;
 
-(defun agda2-annotate (goals)
+(defun agda2-highlight-reload nil
+  "Loads precomputed syntax highlighting info for the current buffer.
+If there is any to load."
+  (let ((highlighting (make-temp-file "agda2-mode")))
+    (unwind-protect
+        (progn
+          (agda2-go highlighting nil t
+                    "cmd_write_highlighting_info"
+                    (agda2-string-quote (buffer-file-name))
+                    (agda2-string-quote highlighting)))
+      (delete-file highlighting))))
+
+(defun agda2-annotate ()
   "Annotates the goals in the current buffer with text properties.
-The goal numbers should be given by GOALS (in the order they
-appear in the buffer)."
+Note that this function should be run /after/ syntax highlighting
+information has been loaded, because the two highlighting
+mechanisms interact in unfortunate ways."
   (agda2-forget-all-goals)
-  (agda2-let (stk top)
+  (agda2-let
+      (stk
+       top
+       (goals (agda2-ask "cmd_goals"
+                         (agda2-string-quote (buffer-file-name)))))
       ((delims() (re-search-forward "[?]\\|[{][-!]\\|[-!][}]\\|--" nil t))
        (is-lone-questionmark ()
           (save-excursion
@@ -814,17 +920,20 @@ text properties."
   (let(ys)(while xs (push (pop xs) ys)(push sep ys))(pop ys)(nreverse ys)))
 
 (defun agda2-goal-Range (o)
-  "Range of goal overlay O."
+  "The Haskell Range of goal overlay O."
   (format "(Range [Interval %s %s])"
           (agda2-mkPos (+ (overlay-start o) 2))
           (agda2-mkPos (- (overlay-end   o) 2))))
 
 (defun agda2-mkPos (&optional p)
-  "Position value of P or point."
+  "The Haskell Position corresponding to P or `point'."
   (save-excursion
     (if p (goto-char p))
-    (format "(Pn \"%s\" %d %d %d)" (buffer-file-name)
-            (point) (count-lines (point-min) (point)) (1+ (current-column)))))
+    (format "(Pn (Just (mkAbsolute %s)) %d %d %d)"
+            (agda2-string-quote (file-truename (buffer-file-name)))
+            (point)
+            (count-lines (point-min) (point))
+            (1+ (current-column)))))
 
 (defun agda2-char-quote (c)
   "Convert character C to the notation used in Haskell strings.
@@ -1022,10 +1131,10 @@ invoked."
 With prefix argument, turn on display of implicit arguments if
 the argument is a positive number, otherwise turn it off."
   (interactive "P")
-  (cond ((eq arg nil)       (agda2-go t "toggleImplicitArgs"))
+  (cond ((eq arg nil)       (agda2-go t t t "toggleImplicitArgs"))
         ((and (numberp arg)
-              (> arg 0))    (agda2-go t "showImplicitArgs" "True"))
-        (t                  (agda2-go t "showImplicitArgs" "False"))))
+              (> arg 0))    (agda2-go t t t "showImplicitArgs" "True"))
+        (t                  (agda2-go t t t "showImplicitArgs" "False"))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;
