@@ -3,21 +3,23 @@
 -- | Generates data used for precise syntax highlighting.
 
 module Agda.Interaction.Highlighting.Generate
-  ( TypeCheckingState(..)
-  , generateSyntaxInfo
+  ( generateSyntaxInfo
   , generateErrorInfo
   , Agda.Interaction.Highlighting.Generate.tests
   )
   where
 
+import Agda.Interaction.FindFile
 import Agda.Interaction.Highlighting.Precise hiding (tests)
 import Agda.Interaction.Highlighting.Range   hiding (tests)
+import qualified Agda.TypeChecking.Errors as E
 import Agda.TypeChecking.MetaVars (isBlockedTerm)
 import Agda.TypeChecking.Monad
   hiding (MetaInfo, Primitive, Constructor, Record, Function, Datatype)
 import qualified Agda.TypeChecking.Monad as M
 import qualified Agda.TypeChecking.Reduce as R
 import qualified Agda.Syntax.Abstract as A
+import qualified Agda.Syntax.Common as SC
 import qualified Agda.Syntax.Concrete as C
 import qualified Agda.Syntax.Info as SI
 import qualified Agda.Syntax.Internal as I
@@ -50,63 +52,86 @@ import Agda.Utils.Impossible
 #include "../../undefined.h"
 
 -- | Generates syntax highlighting information for an error,
--- represented as a range and a string. The range is first completed
--- so that there are no gaps in it.
+-- represented as a range and an optional string. The error range is
+-- completed so that there are no gaps in it.
+--
+-- Nothing is generated unless the file name component of the range is
+-- defined.
 
-generateErrorInfo :: P.Range -> String -> File
+generateErrorInfo :: P.Range -> Maybe String -> Maybe HighlightingInfo
 generateErrorInfo r s =
+  case P.rStart r of
+    Nothing                                          -> Nothing
+    Just (P.Pn { P.srcFile = Nothing })              -> Nothing
+    Just (P.Pn { P.srcFile = Just f, P.posPos = p }) ->
+      Just $ compress $ generateErrorFile r s
+
+-- | Generates syntax highlighting information for an error,
+-- represented as a range and an optional string. The error range is
+-- completed so that there are no gaps in it.
+
+generateErrorFile :: P.Range -> Maybe String -> File
+generateErrorFile r s =
   several (rToR $ P.continuousPerLine r)
-          (mempty { otherAspects = [Error], note = Just s })
-
--- | Has typechecking been done yet?
-
-data TypeCheckingState = TypeCheckingDone | TypeCheckingNotDone
-  deriving (Show, Eq)
+          (mempty { otherAspects = [Error]
+                  , note         = s
+                  })
 
 -- | Generates syntax highlighting information.
 
 generateSyntaxInfo
-  :: FilePath               -- ^ The module to highlight.
-  -> TypeCheckingState      -- ^ Has it been type checked?
+  :: AbsolutePath           -- ^ The module to highlight.
+  -> Maybe TCErr            -- ^ 'Nothing' if the module has been
+                            --   successfully type checked (perhaps
+                            --   with warnings), otherwise the
+                            --   offending error.
+                            --
+                            --   Precondition: The range of the error
+                            --   must match the file name given in the
+                            --   previous argument.
   -> CA.TopLevelInfo        -- ^ The abstract syntax of the module.
   -> [([A.QName], [Range])] -- ^ Functions which failed to termination
                             --   check (grouped if they are mutual),
                             --   along with ranges for problematic
                             --   call sites.
   -> TCM HighlightingInfo
-generateSyntaxInfo file tcs top termErrs =
+generateSyntaxInfo file mErr top termErrs =
   M.withScope_ (CA.insideScope top) $ M.ignoreAbstractMode $ do
-    modMap <- sourceToModule file (CA.topLevelModuleName top)
+    modMap <- sourceToModule
     tokens <- liftIO $ Pa.parseFile' Pa.tokensParser file
-    kinds  <- nameKinds tcs decls
+    kinds  <- nameKinds mErr decls
     let nameInfo = mconcat $ map (generate modMap file kinds)
                                  (Fold.toList names)
     -- Constructors are only highlighted after type checking, since they
     -- can be overloaded.
-    constructorInfo <-
-      if tcs == TypeCheckingNotDone
-         then return mempty
-         else generateConstructorInfo modMap file kinds decls
-    metaInfo <- if tcs == TypeCheckingNotDone
-                   then return mempty
-                   else computeUnsolvedMetaWarnings
+    constructorInfo <- case mErr of
+      Nothing -> generateConstructorInfo modMap file kinds decls
+      Just _  -> return mempty
+    metaInfo <- case mErr of
+      Nothing -> computeUnsolvedMetaWarnings
+      Just _  -> return mempty
+    errorInfo <- case mErr of
+      Nothing -> return mempty
+      Just e  -> let r = P.getRange e in
+        case P.rStart r of
+          Just p | P.srcFile p == Just file ->
+            generateErrorFile r . Just <$> E.prettyError e
+          _ -> __IMPOSSIBLE__
     -- theRest needs to be placed before nameInfo here since record
     -- field declarations contain QNames. constructorInfo also needs
     -- to be placed before nameInfo since, when typechecking is done,
     -- constructors are included in both lists. Finally tokInfo is
     -- placed last since token highlighting is more crude than the
     -- others.
-    return $ HighlightingInfo
-               { source = file
-               , info   = compress $
-                            mconcat [ constructorInfo
-                                    , theRest modMap
-                                    , nameInfo
-                                    , metaInfo
-                                    , termInfo
-                                    , tokInfo tokens
-                                    ]
-               }
+    return $ compress $ mconcat
+      [ errorInfo
+      , constructorInfo
+      , theRest modMap
+      , nameInfo
+      , metaInfo
+      , termInfo
+      , tokInfo tokens
+      ]
   where
     decls = CA.topLevelDecls top
 
@@ -122,6 +147,7 @@ generateSyntaxInfo file tcs top termErrs =
       tokenToFile (T.TokKeyword T.KwForall i)      = aToF Symbol (P.getRange i)
       tokenToFile (T.TokKeyword _ i)               = aToF Keyword (P.getRange i)
       tokenToFile (T.TokSymbol  _ i)               = aToF Symbol (P.getRange i)
+      tokenToFile (T.TokLiteral (L.LitLevel  r _)) = aToF Number r
       tokenToFile (T.TokLiteral (L.LitInt    r _)) = aToF Number r
       tokenToFile (T.TokLiteral (L.LitFloat  r _)) = aToF Number r
       tokenToFile (T.TokLiteral (L.LitString r _)) = aToF String r
@@ -211,10 +237,10 @@ generateSyntaxInfo file tcs top termErrs =
       getPattern _             = mempty
 
       getFieldDecl :: A.Definition -> File
-      getFieldDecl (A.RecDef _ _ _ _ fs) = Fold.foldMap extractField fs
+      getFieldDecl (A.RecDef _ _ _ _ _ fs) = Fold.foldMap extractField fs
         where
         extractField (A.ScopedDecl _ ds) = Fold.foldMap extractField ds
-        extractField (A.Field _ x _)     = field (concreteQualifier x)
+        extractField (A.Field _ _ x _)   = field (concreteQualifier x)
                                                  (concreteBase x)
         extractField _                   = mempty
       getFieldDecl _                   = mempty
@@ -234,14 +260,15 @@ type NameKinds = A.QName -> Maybe NameKind
 
 -- | Builds a 'NameKinds' function.
 
-nameKinds :: TypeCheckingState
+nameKinds :: Maybe TCErr  -- ^ 'Nothing' if type checking completed
+                          --   successfully.
           -> [A.Declaration]
           -> TCM NameKinds
-nameKinds tcs decls = do
+nameKinds mErr decls = do
   imported <- fix . stImports <$> get
-  local    <- case tcs of
-    TypeCheckingDone    -> fix . stSignature <$> get
-    TypeCheckingNotDone -> return $
+  local    <- case mErr of
+    Nothing -> fix . stSignature <$> get
+    Just _  -> return $
       -- Traverses the syntax tree and constructs a map from qualified
       -- names to name kinds. TODO: Handle open public.
       everything' union (Map.empty `mkQ` getDef `extQ` getDecl) decls
@@ -273,17 +300,22 @@ nameKinds tcs decls = do
   getAxiomName _               = __IMPOSSIBLE__
 
   getDef :: A.Definition -> Map A.QName NameKind
-  getDef (A.FunDef  _ q _)      = Map.singleton q Function
-  getDef (A.DataDef _ q i _ cs) = Map.singleton q Datatype `union`
-                                  (Map.unions $
-                                   map (\q -> Map.singleton q (Constructor i)) $
-                                   map getAxiomName cs)
-  getDef (A.RecDef  _ q _ _ _)  = Map.singleton q Record
-  getDef (A.ScopedDef {})       = Map.empty
+  getDef (A.FunDef  _ q _)       = Map.singleton q Function
+  getDef (A.DataDef _ q i _ cs)  = Map.singleton q Datatype `union`
+                                   (Map.unions $
+                                    map (\q -> Map.singleton q (Constructor i)) $
+                                    map getAxiomName cs)
+  getDef (A.RecDef  _ q c _ _ _) = Map.singleton q Record `union`
+                                   case c of
+                                     Nothing -> Map.empty
+                                     Just (A.Axiom _ q _) ->
+                                       Map.singleton q (Constructor SC.Inductive)
+                                     Just _ -> __IMPOSSIBLE__
+  getDef (A.ScopedDef {})        = Map.empty
 
   getDecl :: A.Declaration -> Map A.QName NameKind
   getDecl (A.Axiom _ q _)     = Map.singleton q Postulate
-  getDecl (A.Field _ q _)     = Map.singleton q Field
+  getDecl (A.Field _ _ q _)   = Map.singleton q Field
   getDecl (A.Primitive _ q _) = Map.singleton q Primitive
   getDecl (A.Definition {})   = Map.empty
   getDecl (A.Section {})      = Map.empty
@@ -302,7 +334,7 @@ nameKinds tcs decls = do
 
 generateConstructorInfo
   :: SourceToModule  -- ^ Maps source file paths to module names.
-  -> FilePath        -- ^ The module to highlight.
+  -> AbsolutePath    -- ^ The module to highlight.
   -> NameKinds
   -> [A.Declaration]
   -> TCM File
@@ -335,7 +367,7 @@ generateConstructorInfo modMap file kinds decls = do
   getConstructor :: I.Term -> TCM (Seq A.QName)
   getConstructor (I.Con q _) = return $ Seq.singleton q
   getConstructor (I.Def c _)
-    | fmap P.srcFile (P.rStart (P.getRange c)) == Just file
+    | fmap P.srcFile (P.rStart (P.getRange c)) == Just (Just file)
                              = retrieveCoconstructor c
   getConstructor _           = return Seq.empty
 
@@ -347,14 +379,23 @@ generateConstructorInfo modMap file kinds decls = do
   retrieveCoconstructor c = do
     def <- getConstInfo c
     case defDelayed def of
-      NotDelayed -> return Seq.empty  -- not a coconstructor
-      Delayed -> case defClauses def of
-        [I.Clause{ I.clauseBody = body}] -> case getRHS body of
-          Just (I.Con c args) -> do
-            s <- everything' (liftM2 (><)) query args
-            return $ Seq.singleton c >< s
-          _                   -> return Seq.empty
-        _ -> return Seq.empty
+      -- Not a coconstructor.
+      NotDelayed -> return Seq.empty
+
+      Delayed -> do
+        clauses <- R.instantiateFull $ defClauses def
+        case clauses of
+          [I.Clause{ I.clauseBody = body }] -> case getRHS body of
+            Just (I.Con c args) -> do
+              s <- everything' (liftM2 (><)) query args
+              return $ Seq.singleton c >< s
+
+            -- The meta variable could not be instantiated.
+            Just (I.MetaV {})   -> return Seq.empty
+
+            _                   -> __IMPOSSIBLE__
+
+          _ -> __IMPOSSIBLE__
     where
       getRHS (I.Body v)   = Just v
       getRHS I.NoBody     = Nothing
@@ -382,7 +423,7 @@ computeUnsolvedMetaWarnings = do
 
 generate :: SourceToModule
             -- ^ Maps source file paths to module names.
-         -> FilePath
+         -> AbsolutePath
             -- ^ The module to highlight.
          -> NameKinds
          -> A.AmbiguousQName
@@ -404,7 +445,7 @@ generate modMap file kinds (A.AmbQ qs) =
 
 nameToFile :: SourceToModule
               -- ^ Maps source file paths to module names.
-           -> FilePath
+           -> AbsolutePath
               -- ^ The file name of the current module. Used for
               -- consistency checking.
            -> [C.Name]
@@ -421,7 +462,7 @@ nameToFile :: SourceToModule
            -> File
 nameToFile modMap file xs x m mR =
   -- Make sure that we don't get any funny ranges.
-  if all (== file) $ catMaybes $
+  if all (== Just file) $ catMaybes $
      map (fmap P.srcFile . P.rStart . P.getRange) (x : xs) then
     several rs' ((m isOp) { definitionSite = mFilePos })
    else
@@ -431,15 +472,15 @@ nameToFile modMap file xs x m mR =
   rs'        = rs ++ concatMap (fst . getRanges) xs
   mFilePos   = do
     r <- mR
-    P.Pn { P.srcFile = f, P.posPos = p } <- P.rStart r
+    P.Pn { P.srcFile = Just f, P.posPos = p } <- P.rStart r
     mod <- Map.lookup f modMap
-    return (toStrings mod, f, toInteger p)
+    return (mod, toInteger p)
 
 -- | A variant of 'nameToFile' for qualified abstract names.
 
 nameToFileA :: SourceToModule
                -- ^ Maps source file paths to module names.
-            -> FilePath
+            -> AbsolutePath
                -- ^ The file name of the current module. Used for
                -- consistency checking.
             -> A.QName
@@ -461,23 +502,6 @@ nameToFileA modMap file x include m =
 concreteBase      = A.nameConcrete . A.qnameName
 concreteQualifier = map A.nameConcrete . A.mnameToList . A.qnameModule
 bindingSite       = A.nameBindingSite . A.qnameName
-toStrings         = map show . A.mnameToList
-
--- | Maps source file names to the corresponding top-level module
--- names.
-
-type SourceToModule = Map FilePath A.ModuleName
-
-sourceToModule
-  :: FilePath            -- ^ The current source file.
-  -> A.ModuleName        -- ^ The current top-level module name.
-  -> TCM SourceToModule
-sourceToModule file mod =
-  Map.fromList .
-  (:) (file, mod) .
-  map (\(m, (i, _)) -> (source $ M.iHighlighting i, m)) .
-  Map.toList <$>
-  M.getVisitedModules
 
 -- | Like 'everything', but modified so that it does not descend into
 -- everything.
