@@ -4,7 +4,6 @@ module Agda.Syntax.Concrete.Definitions
     ( NiceDeclaration(..)
     , NiceDefinition(..)
     , NiceConstructor, NiceTypeSignature
-    , NiceField
     , Clause(..)
     , DeclarationException(..)
     , Nice, runNice
@@ -12,13 +11,14 @@ module Agda.Syntax.Concrete.Definitions
     , notSoNiceDeclarations
     ) where
 
-import Control.Exception
 import Control.Applicative
 
 import Data.Generics (Data, Typeable)
 import qualified Data.Map as Map
 import Control.Monad.Error
 import Data.List
+import Data.Maybe
+import qualified Data.Traversable as Trav
 
 import Agda.Syntax.Concrete
 import Agda.Syntax.Common
@@ -41,7 +41,7 @@ import Agda.Utils.Impossible
 -}
 data NiceDeclaration
 	= Axiom Range Fixity Access IsAbstract Name Expr
-        | NiceField Range Fixity Access IsAbstract Name Expr
+        | NiceField Range Fixity Access IsAbstract Hiding Name Expr
 	| PrimitiveFunction Range Fixity Access IsAbstract Name Expr
 	| NiceDef Range [Declaration] [NiceTypeSignature] [NiceDefinition]
 	    -- ^ A bunch of mutually recursive functions\/datatypes.
@@ -58,12 +58,13 @@ data NiceDeclaration
 data NiceDefinition
 	= FunDef  Range [Declaration] Fixity Access IsAbstract Name [Clause]
 	| DataDef Range Induction Fixity Access IsAbstract Name [LamBinding] [NiceConstructor]
-	| RecDef Range Fixity Access IsAbstract Name [LamBinding] [NiceField]
+	| RecDef Range Fixity Access IsAbstract Name (Maybe NiceConstructor) [LamBinding] [NiceDeclaration]
+          -- ^ The 'NiceConstructor' has a dummy type field (the
+          --   record constructor type has not been computed yet).
     deriving (Typeable, Data)
 
 -- | Only 'Axiom's.
 type NiceConstructor = NiceTypeSignature
-type NiceField = NiceTypeSignature
 
 -- | Only 'Axiom's.
 type NiceTypeSignature	= NiceDeclaration
@@ -95,7 +96,7 @@ instance HasRange DeclarationException where
 
 instance HasRange NiceDeclaration where
     getRange (Axiom r _ _ _ _ _)	       = r
-    getRange (NiceField r _ _ _ _ _)	       = r
+    getRange (NiceField r _ _ _ _ _ _)	       = r
     getRange (NiceDef r _ _ _)		       = r
     getRange (NiceModule r _ _ _ _ _)	       = r
     getRange (NiceModuleMacro r _ _ _ _ _ _ _) = r
@@ -107,7 +108,7 @@ instance HasRange NiceDeclaration where
 instance HasRange NiceDefinition where
   getRange (FunDef r _ _ _ _ _ _)    = r
   getRange (DataDef r _ _ _ _ _ _ _) = r
-  getRange (RecDef r _ _ _ _ _ _)    = r
+  getRange (RecDef r _ _ _ _ _ _ _)  = r
 
 instance Error DeclarationException where
   noMsg  = strMsg ""
@@ -132,7 +133,7 @@ instance Show DeclarationException where
     [text $ decl nd] ++ pwords "are not allowed in mutual blocks"
     where
       decl (Axiom _ _ _ _ _ _)		     = "Postulates"
-      decl (NiceField _ _ _ _ _ _)           = "Fields"
+      decl (NiceField _ _ _ _ _ _ _)         = "Fields"
       decl (NiceDef _ _ _ _)		     = "Record types"
       decl (NiceModule _ _ _ _ _ _)	     = "Modules"
       decl (NiceModuleMacro _ _ _ _ _ _ _ _) = "Modules"
@@ -172,12 +173,12 @@ niceDeclarations ds = do
 	declaredNames :: Declaration -> [Name]
 	declaredNames d = case d of
 	  TypeSig x _				       -> [x]
-          Field x _                                    -> [x]
-	  FunClause (LHS p [] _) _ _
+          Field _ x _                                  -> [x]
+	  FunClause (LHS p [] _ _) _ _
             | IdentP (QName x) <- noSingletonRawAppP p -> [x]
 	  FunClause{}				       -> []
 	  Data _ _ x _ _ cs			       -> x : concatMap declaredNames cs
-	  Record _ x _ _ _			       -> [x]
+	  Record _ x c _ _ _			       -> x : maybeToList c
 	  Infix _ _				       -> []
 	  Mutual _ ds				       -> concatMap declaredNames ds
 	  Abstract _ ds				       -> concatMap declaredNames ds
@@ -207,7 +208,7 @@ niceDeclarations ds = do
 			  d <- mkFunDef fixs x (Just t) ds0
                           return $ d : ds1
 
-		cl@(FunClause lhs@(LHS p [] _) _ _)
+		cl@(FunClause lhs@(LHS p [] _ _) _ _)
                   | IdentP (QName x) <- noSingletonRawAppP p
                                   -> do
 		      ds <- nice fixs ds
@@ -218,9 +219,13 @@ niceDeclarations ds = do
 		_   -> liftM2 (++) nds (nice fixs ds)
 		    where
 			nds = case d of
-                            Field x t               -> return $ niceAxioms fixs [ Field x t ]
+                            Field h x t             -> return $ niceAxioms fixs [ Field h x t ]
 			    Data   r ind x tel t cs -> dataOrRec (flip DataDef ind) niceAx r x tel t cs
-			    Record r x tel t cs     -> dataOrRec RecDef (const niceDeclarations) r x tel t cs
+			    Record r x c tel t cs   -> do
+                              let dummyType = Prop noRange
+                                  c'        = (\c -> niceAxiom fixs (TypeSig c dummyType)) <$> c
+                              dataOrRec (\x1 x2 x3 x4 x5 -> RecDef x1 x2 x3 x4 x5 c')
+                                        (const niceDeclarations) r x tel t cs
 			    Mutual r ds -> do
 			      d <- mkMutual r [d] =<< niceFix fixs ds
 			      return [d]
@@ -280,16 +285,14 @@ niceDeclarations ds = do
         niceAx fixs ds = return $ niceAxioms fixs ds
 
 	niceAxioms :: Map.Map Name Fixity -> [TypeSignature] -> [NiceDeclaration]
-	niceAxioms fixs ds = nice ds
-	    where
-		nice [] = []
-		nice (d@(TypeSig x t) : ds) =
-		    Axiom (getRange d) (fixity x fixs) PublicAccess ConcreteDef x t
-		    : nice ds
-		nice (d@(Field x t) : ds) =
-		    NiceField (getRange d) (fixity x fixs) PublicAccess ConcreteDef x t
-		    : nice ds
-		nice _ = __IMPOSSIBLE__
+	niceAxioms fixs ds = map (niceAxiom fixs) ds
+
+        niceAxiom :: Map.Map Name Fixity -> TypeSignature -> NiceDeclaration
+        niceAxiom fixs d@(TypeSig x t) =
+            Axiom (getRange d) (fixity x fixs) PublicAccess ConcreteDef x t
+        niceAxiom fixs d@(Field h x t) =
+            NiceField (getRange d) (fixity x fixs) PublicAccess ConcreteDef h x t
+        niceAxiom _ _ = __IMPOSSIBLE__
 
 	toPrim :: NiceDeclaration -> NiceDeclaration
 	toPrim (Axiom r f a c x t) = PrimitiveFunction r f a c x t
@@ -313,19 +316,19 @@ niceDeclarations ds = do
 
         expandEllipsis :: [Declaration] -> [Declaration]
         expandEllipsis [] = []
-        expandEllipsis (d@(FunClause (Ellipsis _ _ _) _ _) : ds) =
+        expandEllipsis (d@(FunClause Ellipsis{} _ _) : ds) =
           d : expandEllipsis ds
-        expandEllipsis (d@(FunClause lhs@(LHS p ps _) _ _) : ds) =
+        expandEllipsis (d@(FunClause lhs@(LHS p ps _ _) _ _) : ds) =
           d : expand p ps ds
           where
             expand _ _ [] = []
-            expand p ps (FunClause (Ellipsis _ ps' []) rhs wh : ds) =
-              FunClause (LHS p (ps ++ ps') []) rhs wh : expand p ps ds
-            expand p ps (FunClause (Ellipsis _ ps' es) rhs wh : ds) =
-              FunClause (LHS p (ps ++ ps') es) rhs wh : expand p (ps ++ ps') ds
-            expand p ps (d@(FunClause (LHS _ _ []) _ _) : ds) =
+            expand p ps (FunClause (Ellipsis _ ps' eqs []) rhs wh : ds) =
+              FunClause (LHS p (ps ++ ps') eqs []) rhs wh : expand p ps ds
+            expand p ps (FunClause (Ellipsis _ ps' eqs es) rhs wh : ds) =
+              FunClause (LHS p (ps ++ ps') eqs es) rhs wh : expand p (ps ++ ps') ds
+            expand p ps (d@(FunClause (LHS _ _ _ []) _ _) : ds) =
               d : expand p ps ds
-            expand _ _ (d@(FunClause (LHS p ps (_ : _)) _ _) : ds) =
+            expand _ _ (d@(FunClause (LHS p ps _ (_ : _)) _ _) : ds) =
               d : expand p ps ds
             expand _ _ (_ : ds) = __IMPOSSIBLE__
         expandEllipsis (_ : ds) = __IMPOSSIBLE__
@@ -334,9 +337,9 @@ niceDeclarations ds = do
         -- Turn function clauses into nice function clauses.
         mkClauses :: Name -> [Declaration] -> Nice [Clause]
         mkClauses _ [] = return []
-        mkClauses x (FunClause lhs@(LHS _ _ []) rhs wh : cs) =
+        mkClauses x (FunClause lhs@(LHS _ _ _ []) rhs wh : cs) =
           (Clause x lhs rhs wh [] :) <$> mkClauses x cs
-        mkClauses x (FunClause lhs@(LHS _ ps es) rhs wh : cs) = do
+        mkClauses x (FunClause lhs@(LHS _ ps _ es) rhs wh : cs) = do
           when (null with) $ throwError $ MissingWithClauses x
           wcs <- mkClauses x with
           (Clause x lhs rhs wh wcs :) <$> mkClauses x cs'
@@ -346,18 +349,19 @@ niceDeclarations ds = do
             -- A clause is a subclause if the number of with-patterns is
             -- greater or equal to the current number of with-patterns plus the
             -- number of with arguments.
-            subClause (FunClause (LHS _ ps' _) _ _)      = length ps' >= length ps + length es
-            subClause (FunClause (Ellipsis _ ps' _) _ _) = True
+            subClause (FunClause (LHS _ ps' _ _) _ _)      =
+              length ps' >= length ps + length es
+            subClause (FunClause (Ellipsis _ ps' _ _) _ _) = True
             subClause _                                  = __IMPOSSIBLE__
-        mkClauses x (FunClause lhs@(Ellipsis _ _ _) rhs wh : cs) =
+        mkClauses x (FunClause lhs@Ellipsis{} rhs wh : cs) =
           (Clause x lhs rhs wh [] :) <$> mkClauses x cs   -- Will result in an error later.
         mkClauses _ _ = __IMPOSSIBLE__
 
 	noSingletonRawAppP (RawAppP _ [p]) = noSingletonRawAppP p
 	noSingletonRawAppP p		   = p
 
-        isFunClauseOf x (FunClause (Ellipsis _ _ _) _ _) = True
-	isFunClauseOf x (FunClause (LHS p _ _) _ _) = case noSingletonRawAppP p of
+        isFunClauseOf x (FunClause Ellipsis{} _ _) = True
+	isFunClauseOf x (FunClause (LHS p _ _ _) _ _) = case noSingletonRawAppP p of
 	    IdentP (QName q)	-> x == q
 	    _			-> True
 		-- more complicated lhss must come with type signatures, so we just assume
@@ -389,7 +393,7 @@ niceDeclarations ds = do
 	mkAbstract d =
 	    case d of
 		Axiom r f a _ x e		    -> Axiom r f a AbstractDef x e
-		NiceField r f a _ x e		    -> NiceField r f a AbstractDef x e
+		NiceField r f a _ h x e		    -> NiceField r f a AbstractDef h x e
 		PrimitiveFunction r f a _ x e	    -> PrimitiveFunction r f a AbstractDef x e
 		NiceDef r cs ts ds		    -> NiceDef r cs (map mkAbstract ts)
 								 (map mkAbstractDef ds)
@@ -404,7 +408,7 @@ niceDeclarations ds = do
 		FunDef r ds f a _ x cs      -> FunDef r ds f a AbstractDef x
 						  (map mkAbstractClause cs)
 		DataDef r ind f a _ x ps cs -> DataDef r ind f a AbstractDef x ps $ map mkAbstract cs
-		RecDef r f a _ x ps cs      -> RecDef r f a AbstractDef x ps $ map mkAbstract cs
+		RecDef r f a _ x c ps cs    -> RecDef r f a AbstractDef x (mkAbstract <$> c) ps $ map mkAbstract cs
 
 	mkAbstractClause (Clause x lhs rhs wh with) =
 	    Clause x lhs rhs (mkAbstractWhere wh) (map mkAbstractClause with)
@@ -417,7 +421,7 @@ niceDeclarations ds = do
 	mkPrivate d =
 	    case d of
 		Axiom r f _ a x e		    -> Axiom r f PrivateAccess a x e
-		NiceField r f _ a x e		    -> NiceField r f PrivateAccess a x e
+		NiceField r f _ a h x e		    -> NiceField r f PrivateAccess a h x e
 		PrimitiveFunction r f _ a x e	    -> PrimitiveFunction r f PrivateAccess a x e
 		NiceDef r cs ts ds		    -> NiceDef r cs (map mkPrivate ts)
 								    (map mkPrivateDef ds)
@@ -432,7 +436,7 @@ niceDeclarations ds = do
 		FunDef r ds f _ a x cs      -> FunDef r ds f PrivateAccess a x
 						  (map mkPrivateClause cs)
 		DataDef r ind f _ a x ps cs -> DataDef r ind f PrivateAccess a x ps (map mkPrivate cs)
-		RecDef r f _ a x ps cs      -> RecDef r f PrivateAccess a x ps cs
+		RecDef r f _ a x c ps cs    -> RecDef r f PrivateAccess a x (mkPrivate <$> c) ps cs
 
 	mkPrivateClause (Clause x lhs rhs wh with) =
 	    Clause x lhs rhs (mkPrivateWhere wh) (map mkPrivateClause with)
@@ -466,7 +470,7 @@ notSoNiceDeclarations :: [NiceDeclaration] -> [Declaration]
 notSoNiceDeclarations = concatMap notNice
   where
     notNice (Axiom _ _ _ _ x e)                   = [TypeSig x e]
-    notNice (NiceField _ _ _ _ x e)               = [Field x e]
+    notNice (NiceField _ _ _ _ h x e)             = [Field h x e]
     notNice (PrimitiveFunction r _ _ _ x e)       = [Primitive r [TypeSig x e]]
     notNice (NiceDef _ ds _ _)                    = ds
     notNice (NiceModule r _ _ x tel ds)           = [Module r x tel ds]
