@@ -29,6 +29,7 @@ import Data.Map (Map)
 import qualified Data.Set as Set
 import Data.Set (Set)
 import Data.List as List
+import qualified Data.Traversable as Trav
 
 import Agda.Syntax.Common
 import Agda.Syntax.Position
@@ -44,7 +45,7 @@ import Agda.TypeChecking.Monad.State (getScope)
 import Agda.TypeChecking.Monad.Base  (MonadTCM)
 
 import Agda.Utils.Maybe
-import Agda.Utils.Monad
+import Agda.Utils.Monad hiding (bracket)
 import Agda.Utils.Tuple
 import Agda.Utils.Suffix
 
@@ -406,8 +407,11 @@ instance ToConcrete A.TypedBinding C.TypedBinding where
 instance ToConcrete LetBinding [C.Declaration] where
     bindToConcrete (LetBind i x t e) ret =
         bindToConcrete x $ \x ->
-        do  (t,(e, [], [])) <- toConcrete (t, A.RHS e)
-            ret [C.TypeSig x t, C.FunClause (C.LHS (C.IdentP $ C.QName x) [] []) e C.NoWhere]
+        do  (t,(e, [], [], [])) <- toConcrete (t, A.RHS e)
+            ret [ C.TypeSig x t
+                , C.FunClause (C.LHS (C.IdentP $ C.QName x) [] [] [])
+                              e C.NoWhere
+                ]
     bindToConcrete (LetApply i x tel y es _ _) ret = do
       x  <- unsafeQNameToName <$> toConcrete x
       y  <- toConcrete y
@@ -428,15 +432,22 @@ instance ToConcrete LetBinding [C.Declaration] where
 instance ToConcrete [A.Declaration] [C.Declaration] where
     toConcrete ds = concat <$> mapM toConcrete ds
 
-instance ToConcrete A.RHS (C.RHS, [C.Expr], [C.Declaration]) where
+instance ToConcrete A.RHS (C.RHS, [C.Expr], [C.Expr], [C.Declaration]) where
     toConcrete (A.RHS e) = do
       e <- toConcrete e
-      return (C.RHS e, [], [])
-    toConcrete A.AbsurdRHS = return (C.AbsurdRHS, [], [])
+      return (C.RHS e, [], [], [])
+    toConcrete A.AbsurdRHS = return (C.AbsurdRHS, [], [], [])
     toConcrete (A.WithRHS _ es cs) = do
       es <- toConcrete es
       cs <- toConcrete cs
-      return (C.AbsurdRHS, es, concat cs)
+      return (C.AbsurdRHS, [], es, concat cs)
+    toConcrete (A.RewriteRHS _ eqs rhs wh) = do
+      wh <- toConcrete wh
+      (rhs, eqs', es, whs) <- toConcrete rhs
+      unless (null eqs')
+        __IMPOSSIBLE__
+      eqs <- toConcrete eqs
+      return (rhs, eqs, es, wh ++ whs)
 
 data TypeAndDef = TypeAndDef A.TypeSignature A.Definition
 
@@ -468,17 +479,20 @@ instance ToConcrete TypeAndDef [C.Declaration] where
       mkTel n (A.Pi _ b t) = (b++) -*- id $ mkTel (n - 1) t
       mkTel _ _            = __IMPOSSIBLE__
 
-  toConcrete (TypeAndDef (Axiom _ x t) (RecDef  i _ bs _ cs)) =
+  toConcrete (TypeAndDef (Axiom _ x t) (RecDef  i _ c bs _ cs)) =
     withAbstractPrivate i $
     bindToConcrete tel $ \tel' -> do
       t'       <- toConcreteCtx TopCtx t0
       (x',cs') <- (unsafeQNameToName -*- id) <$> toConcrete (x, map Constr cs)
-      return [ C.Record (getRange i) x' tel' t' cs' ]
+      c'       <- Trav.mapM (\d -> unsafeQNameToName <$>
+                                     toConcrete (axiomName d)) c
+      return [ C.Record (getRange i) x' c' tel' t' cs' ]
     where
       (tel, t0) = mkTel (length bs) t
-      mkTel 0 t            = ([], t)
-      mkTel n (A.Pi _ b t) = (b++) -*- id $ mkTel (n - 1) t
-      mkTel _ _            = __IMPOSSIBLE__
+      mkTel n (A.ScopedExpr _ t) = mkTel n t
+      mkTel 0 t                  = ([], t)
+      mkTel n (A.Pi _ b t)       = (b++) -*- id $ mkTel (n - 1) t
+      mkTel _ _                  = __IMPOSSIBLE__
 
   toConcrete _ = __IMPOSSIBLE__
 
@@ -489,17 +503,17 @@ instance ToConcrete (Constr A.Constructor) C.Declaration where
     x' <- unsafeQNameToName <$> toConcrete x
     t' <- toConcreteCtx TopCtx t
     return $ C.TypeSig x' t'
-  toConcrete _ = __IMPOSSIBLE__
+  toConcrete (Constr d) = head <$> toConcrete d
 
 instance ToConcrete A.Clause [C.Declaration] where
   toConcrete (A.Clause lhs rhs wh) =
-      bindToConcrete lhs $ \(C.LHS p wps _) -> do
-          (rhs', with, wcs) <- toConcreteCtx TopCtx rhs
+      bindToConcrete lhs $ \(C.LHS p wps _ _) -> do
+          (rhs', eqs, with, wcs) <- toConcreteCtx TopCtx rhs
           ds         <- toConcrete wh
           let wh' = case ds of
                 []  -> C.NoWhere
                 _   -> C.AnyWhere ds
-          return $ FunClause (C.LHS p wps with) rhs' wh' : wcs
+          return $ FunClause (C.LHS p wps eqs with) rhs' wh' : wcs
 
 instance ToConcrete A.Declaration [C.Declaration] where
   toConcrete (ScopedDecl scope ds) =
@@ -512,12 +526,12 @@ instance ToConcrete A.Declaration [C.Declaration] where
       t' <- toConcreteCtx TopCtx t
       return [C.Postulate (getRange i) [C.TypeSig x' t']]
 
-  toConcrete (A.Field i x t) = do
+  toConcrete (A.Field i h x t) = do
     x' <- unsafeQNameToName <$> toConcrete x
     withAbstractPrivate i $
       withInfixDecl i x'  $ do
       t' <- toConcreteCtx TopCtx t
-      return [C.Field x' t']
+      return [C.Field h x' t']
 
   toConcrete (A.Primitive i x t) = do
     x' <- unsafeQNameToName <$> toConcrete x
@@ -600,7 +614,7 @@ instance ToConcrete A.LHS C.LHS where
     bindToConcrete (A.LHS i x args wps) ret = do
       bindToConcreteCtx TopCtx (A.DefP info x args) $ \lhs ->
         bindToConcreteCtx TopCtx (noImplicitPats wps) $ \wps ->
-          ret $ C.LHS lhs wps []
+          ret $ C.LHS lhs wps [] []
       where info = PatRange (getRange i)
 
 appBrackets' :: [arg] -> Precedence -> Bool
@@ -670,7 +684,8 @@ recoverOpApp bracket opApp view e mdefault = case view e of
     | all notHidden args  -> do
       let  args' = map (namedThing . unArg) args
       case hd of
-        HdVar n  -> do
+        HdVar n | isNoName n -> mdefault
+                | otherwise  -> do
           x <- toConcrete n
           doCName (nameFixity n) x args'
         HdDef qn -> doQName qn args'
@@ -678,6 +693,8 @@ recoverOpApp bracket opApp view e mdefault = case view e of
     | otherwise -> mdefault
   where
 
+  isNoName x = C.isNoName $ A.nameConcrete x
+
   notHidden (Arg h _) = h == NotHidden
 
   -- qualified names can't use mixfix syntax
@@ -689,6 +706,7 @@ recoverOpApp bracket opApp view e mdefault = case view e of
 
   -- fall-back (wrong number of arguments or no holes)
   doCName _ cn@(C.Name _ xs) es
+    | length xs == 1        = mdefault
     | length es /= numHoles = mdefault
     | List.null es          = mdefault
     where numHoles = length [ () | Hole <- xs ]
@@ -700,7 +718,9 @@ recoverOpApp bracket opApp view e mdefault = case view e of
     , Hole <- last xs = do
         let a1     = head as
             an     = last as
-            as'    = init $ tail as
+            as'    = case as of
+                       as@(_:_:_) -> init $ tail as
+                       _          -> __IMPOSSIBLE__
         e1 <- toConcreteCtx (LeftOperandCtx fixity) a1
         es <- mapM (toConcreteCtx InsideOperandCtx) as'
         en <- toConcreteCtx (RightOperandCtx fixity) an
@@ -711,7 +731,9 @@ recoverOpApp bracket opApp view e mdefault = case view e of
   doCName fixity cn@(C.Name _ xs) as
     | Hole <- last xs = do
         let an  = last as
-            as' = init as
+            as' = case as of
+                    as@(_:_) -> init as
+                    _        -> __IMPOSSIBLE__
         es <- mapM (toConcreteCtx InsideOperandCtx) as'
         en <- toConcreteCtx (RightOperandCtx fixity) an
         bracket (opBrackets fixity)
