@@ -30,7 +30,7 @@ import Data.Traversable (mapM)
 import Data.List ((\\), nub)
 import qualified Data.Map as Map
 
-import Agda.Syntax.Concrete as C
+import Agda.Syntax.Concrete as C hiding (topLevelModuleName)
 import Agda.Syntax.Abstract as A
 import Agda.Syntax.Position
 import Agda.Syntax.Common
@@ -42,7 +42,8 @@ import Agda.Syntax.Scope.Base
 import Agda.Syntax.Scope.Monad
 import Agda.Syntax.Strict
 
-import Agda.TypeChecking.Monad.Base (TypeError(..), Call(..), typeError, TCErr(..))
+import Agda.TypeChecking.Monad.Base (TypeError(..), Call(..), typeError,
+                                     TCErr(..), TCErr'(..))
 import Agda.TypeChecking.Monad.Trace (traceCall, traceCallCPS, setCurrentRange)
 import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Monad.Options
@@ -119,9 +120,9 @@ annotateExpr m = do
   return $ ScopedExpr s e
 
 expandEllipsis :: C.Pattern -> [C.Pattern] -> C.Clause -> C.Clause
-expandEllipsis _ _ c@(C.Clause _ (C.LHS _ _ _) _ _ _) = c
-expandEllipsis p ps (C.Clause x (C.Ellipsis _ ps' es) rhs wh wcs) =
-  C.Clause x (C.LHS p (ps ++ ps') es) rhs wh wcs
+expandEllipsis _ _ c@(C.Clause _ C.LHS{} _ _ _) = c
+expandEllipsis p ps (C.Clause x (C.Ellipsis _ ps' eqs es) rhs wh wcs) =
+  C.Clause x (C.LHS p (ps ++ ps') eqs es) rhs wh wcs
 
 -- | Make sure that each variable occurs only once.
 checkPatternLinearity :: [A.Pattern' e] -> ScopeM ()
@@ -151,9 +152,9 @@ recordConstructorType fields = build fs
     notField NiceField{} = False
     notField _           = True
 
-    build (NiceField r f _ _ x e : fs) = C.Pi [C.TypedBindings r NotHidden
-                                                [C.TBind r [BName x f] e]
-                                              ] $ build fs
+    build (NiceField r f _ _ h x e : fs) = C.Pi [C.TypedBindings r h
+                                                  [C.TBind r [BName x f] e]
+                                                ] $ build fs
       where r = getRange x
     build (d : fs)                     = C.Let noRange (notSoNiceDeclarations [d]) $ build fs
     build []                           = C.Prop noRange
@@ -318,7 +319,9 @@ instance ToAbstract PatName APatName where
       (DefinedName d, C.QName x) | DefName == anameKind d -> return $ Left x
       (UnknownName,   C.QName x)                          -> return $ Left x
       (ConstructorName ds, _)                             -> return $ Right ds
-      _                                                   -> fail $ "not a constructor: " ++ show x -- TODO
+      _                                                   ->
+        typeError $ GenericError $
+          "Cannot pattern match on " ++ show x ++ ", because it is not a constructor"
     case z of
       Left x  -> do
         reportSLn "scope.pat" 10 $ "it was a var: " ++ show x
@@ -548,7 +551,7 @@ instance ToAbstract (TopLevel [C.Declaration]) TopLevelInfo where
 
 niceDecls :: [C.Declaration] -> ScopeM [NiceDeclaration]
 niceDecls ds = case runNice $ niceDeclarations ds of
-  Left e   -> throwError $ Exception (getRange e) (show e)
+  Left e   -> throwError $ TCErr Nothing $ Exception (getRange e) (show e)
   Right ds -> return ds
 
 instance ToAbstract [C.Declaration] [A.Declaration] where
@@ -588,7 +591,7 @@ instance ToAbstract LetDef [A.LetBinding] where
 
             _   -> notAValidLetBinding d
         where
-            letToAbstract (C.Clause top clhs@(C.LHS p [] []) (C.RHS rhs) NoWhere []) = do
+            letToAbstract (C.Clause top clhs@(C.LHS p [] [] []) (C.RHS rhs) NoWhere []) = do
                 p    <- parseLHS (Just top) p
                 localToAbstract (snd $ lhsArgs p) $ \args ->
                     do  rhs <- toAbstract rhs
@@ -628,10 +631,10 @@ instance ToAbstract C.Pragma [A.Pragma] where
     toAbstract (C.BuiltinPragma _ b e) = do
         e <- toAbstract e
         return [ A.BuiltinPragma b e ]
-    toAbstract (C.LinePragma _ _ _) = return []
     toAbstract (C.ImportPragma _ i) = do
       addHaskellImport i
       return []
+    toAbstract (C.ImpossiblePragma _) = __IMPOSSIBLE__
 
 -- Only constructor names are bound by definitions.
 instance ToAbstract NiceDefinition Definition where
@@ -667,7 +670,7 @@ instance ToAbstract NiceDefinition Definition where
           conName _ = __IMPOSSIBLE__
 
     -- Record definitions (mucho interesting)
-      C.RecDef r f p a x pars fields ->
+      C.RecDef r f p a x c pars fields ->
         traceCall (ScopeCheckDefinition d) $
         withLocalVars $ do
           pars   <- toAbstract pars
@@ -686,7 +689,8 @@ instance ToAbstract NiceDefinition Definition where
           -- popScope p
           bindModule p x m
           printScope "rec" 15 "record complete"
-          return $ A.RecDef (mkDefInfo x f p a r) x' pars contel afields
+          c' <- mapM (toAbstract . Constr) c
+          return $ A.RecDef (mkDefInfo x f p a r) x' c' pars contel afields
 
 -- The only reason why we return a list is that open declarations disappears.
 -- For every other declaration we get a singleton list.
@@ -704,11 +708,11 @@ instance ToAbstract NiceDeclaration A.Declaration where
       return [ A.Axiom (mkDefInfo x f p a r) y t' ]
 
   -- Fields
-    C.NiceField r f p a x t -> do
+    C.NiceField r f p a h x t -> do
       t' <- toAbstractCtx TopCtx t
       y  <- freshAbstractQName f x
       bindName p DefName x y
-      return [ A.Field (mkDefInfo x f p a r) y t' ]
+      return [ A.Field (mkDefInfo x f p a r) h y t' ]
 
   -- Primitive function
     PrimitiveFunction r f p a x t -> do
@@ -816,8 +820,8 @@ instance ToAbstract (Constr C.NiceDeclaration) A.Declaration where
     toAbstract _ = __IMPOSSIBLE__    -- a constructor is always an axiom
 
 instance ToAbstract C.Clause A.Clause where
-    toAbstract (C.Clause top (C.Ellipsis _ _ _) _ _ _) = fail "bad '...'" -- TODO: errors message
-    toAbstract (C.Clause top lhs@(C.LHS p wps with) rhs wh wcs) = withLocalVars $ do
+    toAbstract (C.Clause top C.Ellipsis{} _ _ _) = fail "bad '...'" -- TODO: errors message
+    toAbstract (C.Clause top lhs@(C.LHS p wps eqs with) rhs wh wcs) = withLocalVars $ do
       let wcs' = map (expandEllipsis p wps) wcs
       lhs' <- toAbstract (LeftHandSide top p wps)
       printLocals 10 "after lhs:"
@@ -825,50 +829,74 @@ instance ToAbstract C.Clause A.Clause where
             NoWhere        -> (Nothing, [])
             AnyWhere ds    -> (Nothing, ds)
             SomeWhere m ds -> (Just m, ds)
-      case whds of
-        [] -> do
-          rhs <- toAbstract =<< toAbstractCtx TopCtx (RightHandSide with wcs' rhs)
+      if not (null eqs)
+        then do
+          rhs <- toAbstract =<< toAbstractCtx TopCtx (RightHandSide eqs with wcs' rhs whds)
           return $ A.Clause lhs' rhs []
-        _       -> do
-          m <- maybe (nameConcrete <$> freshNoName noRange) return whname
-          let acc = maybe PrivateAccess (const PublicAccess) whname  -- unnamed where's are private
-          let tel = []
-          old <- getCurrentModule
-          am  <- toAbstract (NewModuleName m)
-          (scope, ds) <- scopeCheckModule (getRange wh) (C.QName m) am tel whds
-          setScope scope
+        else do
           -- the right hand side is checked inside the module of the local definitions
-          rhs <- toAbstractCtx TopCtx (RightHandSide with wcs' rhs)
-          setCurrentModule old
---        case acc of
---          PublicAccess  -> popScope PublicAccess
---          PrivateAccess -> popScope_  -- unnamed where clauses are not in scope
-          bindModule acc m am
+          (rhs, ds) <- whereToAbstract (getRange wh) whname whds $
+                        toAbstractCtx TopCtx (RightHandSide eqs with wcs' rhs [])
           rhs <- toAbstract rhs
           return $ A.Clause lhs' rhs ds
 
-data RightHandSide = RightHandSide [C.Expr] [C.Clause] C.RHS
+whereToAbstract :: Range -> Maybe C.Name -> [C.Declaration] -> ScopeM a -> ScopeM (a, [A.Declaration])
+whereToAbstract _ _ [] inner = do
+  x <- inner
+  return (x, [])
+whereToAbstract r whname whds inner = do
+  m <- maybe (nameConcrete <$> freshNoName noRange) return whname
+  let acc = maybe PrivateAccess (const PublicAccess) whname  -- unnamed where's are private
+  let tel = []
+  old <- getCurrentModule
+  am  <- toAbstract (NewModuleName m)
+  (scope, ds) <- scopeCheckModule r (C.QName m) am tel whds
+  setScope scope
+  x <- inner
+  setCurrentModule old
+  bindModule acc m am
+  return (x, ds)
+
+data RightHandSide = RightHandSide [C.Expr] [C.Expr] [C.Clause] C.RHS [C.Declaration]
 data AbstractRHS = AbsurdRHS'
                  | WithRHS' [A.Expr] [C.Clause]  -- ^ The with clauses haven't been translated yet
                  | RHS' A.Expr
+                 | RewriteRHS' [A.Expr] AbstractRHS [A.Declaration]
+
+withFunctionName :: String -> ScopeM A.QName
+withFunctionName s = do
+  m <- getCurrentModule
+  NameId i _ <- fresh
+  A.qualify m <$> freshName_ (s ++ show i)
 
 instance ToAbstract AbstractRHS A.RHS where
-  toAbstract AbsurdRHS'       = return A.AbsurdRHS
-  toAbstract (RHS' e)         = return $ A.RHS e
+  toAbstract AbsurdRHS'            = return A.AbsurdRHS
+  toAbstract (RHS' e)              = return $ A.RHS e
+  toAbstract (RewriteRHS' eqs rhs wh) = do
+    auxs <- replicateM (length eqs) $ withFunctionName "rewrite-"
+    rhs  <- toAbstract rhs
+    return $ RewriteRHS auxs eqs rhs wh
   toAbstract (WithRHS' es cs) = do
-    m   <- getCurrentModule
-    -- Hack
-    NameId i _ <- fresh
-    aux <- A.qualify m <$> freshName_ ("aux" ++ show i)
+    aux <- withFunctionName "with-"
     A.WithRHS aux es <$> toAbstract cs
 
 instance ToAbstract RightHandSide AbstractRHS where
-  toAbstract (RightHandSide [] (_ : _) _)        = __IMPOSSIBLE__
-  toAbstract (RightHandSide (_ : _) _ (C.RHS _)) = typeError $ BothWithAndRHS
-  toAbstract (RightHandSide [] [] rhs)           = toAbstract rhs
-  toAbstract (RightHandSide es cs C.AbsurdRHS)   = do
+  toAbstract (RightHandSide eqs@(_:_) es cs rhs wh) = do
+    eqs <- toAbstractCtx TopCtx eqs
+                 -- TODO: remember named where
+    (rhs, ds) <- whereToAbstract (getRange wh) Nothing wh $
+                  toAbstract (RightHandSide [] es cs rhs [])
+    return $ RewriteRHS' eqs rhs ds
+  toAbstract (RightHandSide [] [] (_ : _) _ _)        = __IMPOSSIBLE__
+  toAbstract (RightHandSide [] (_ : _) _ (C.RHS _) _) = typeError $ BothWithAndRHS
+  toAbstract (RightHandSide [] [] [] rhs [])          = toAbstract rhs
+  toAbstract (RightHandSide [] es cs C.AbsurdRHS [])  = do
     es <- toAbstractCtx TopCtx es
     return $ WithRHS' es cs
+  -- TODO: some of these might be possible
+  toAbstract (RightHandSide [] (_ : _) _ C.AbsurdRHS (_ : _)) = __IMPOSSIBLE__
+  toAbstract (RightHandSide [] [] [] (C.RHS _) (_ : _))       = __IMPOSSIBLE__
+  toAbstract (RightHandSide [] [] [] C.AbsurdRHS (_ : _))     = __IMPOSSIBLE__
 
 instance ToAbstract C.RHS AbstractRHS where
     toAbstract C.AbsurdRHS = return $ AbsurdRHS'
