@@ -7,9 +7,11 @@ import Control.Monad
 import Control.Monad.State
 import Control.Monad.Error
 import Data.Generics
-import Data.Traversable hiding (mapM)
+import Data.Traversable hiding (mapM, sequence)
 import Data.List hiding (sort)
+import qualified Data.List as List
 
+import Agda.Syntax.Literal
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.TypeChecking.Monad
@@ -24,6 +26,8 @@ import Agda.TypeChecking.Records
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Injectivity
 import Agda.TypeChecking.SizedTypes
+import Agda.TypeChecking.Monad.Builtin
+import Agda.TypeChecking.Level
 
 import Agda.Utils.Monad
 
@@ -66,10 +70,12 @@ compareTerm cmp a m n =
     proofIrr <- proofIrrelevance
     isSize   <- isSizeType a'
     s        <- reduce $ getSort a'
+    mlvl     <- liftTCM $ (Just <$> primLevel) `catchError` \_ -> return Nothing
     case s of
       Prop | proofIrr -> return []
       _    | isSize   -> compareSizes cmp m n
       _               -> case unEl a' of
+        a | Just a == mlvl && cmp == CmpEq -> equalLevel m n
         Pi a _    -> equalFun (a,a') m n
         Fun a _   -> equalFun (a,a') m n
         MetaV x _ -> do
@@ -198,7 +204,7 @@ compareAtom cmp t m n =
 	    _ | f1@(FunV _ _) <- funView m
 	      , f2@(FunV _ _) <- funView n -> equalFun f1 f2
 
-	    (Sort s1, Sort s2) -> compareSort cmp s1 s2
+	    (Sort s1, Sort s2) -> compareSort CmpEq s1 s2
 
 	    (Lit l1, Lit l2) | l1 == l2 -> return []
 	    (Var i iArgs, Var j jArgs) | i == j -> do
@@ -274,6 +280,13 @@ compareArgs _ _ (_:_) [] = __IMPOSSIBLE__
 compareArgs pols0 a (arg1 : args1) (arg2 : args2) = do
     let (pol, pols) = nextPolarity pols0
     a <- reduce a
+    reportSDoc "tc.conv.args" 30 $
+      sep [ text "compareArgs"
+          , nest 2 $ sep [ prettyTCM (arg1 : args1)
+                         , prettyTCM (arg2 : args2)
+                         ]
+          , nest 2 $ text ":" <+> prettyTCM a
+          ]
     case funView (unEl a) of
 	FunV (Arg _ b) _ -> do
 	    reportSDoc "tc.conv.args" 10 $
@@ -289,16 +302,16 @@ compareArgs pols0 a (arg1 : args1) (arg2 : args2) = do
                             Contravariant -> compareTerm CmpLeq b y x
             cs1 <- cmp (unArg arg1) (unArg arg2)
 	    case (cs1, unEl a) of
-		(_:_, Pi _ c) | 0 `freeIn` absBody c
+                                -- We can safely ignore sort annotations here.
+		(_:_, Pi _ c) | 0 `freeInIgnoringSorts` absBody c
 		    -> do
                         reportSDoc "tc.conv.args" 15 $ sep
                           [ text "aborting compareArgs" <+> parens (text $ show pol)
-                          , nest 2 $ sep
-                              [ parens $ text (show pol)
-                              , prettyTCM arg1
+                          , nest 2 $ fsep
+                              [ prettyTCM arg1
                               , text "~~" <+> prettyTCM arg2
                               , text ":" <+> prettyTCM b
-                              , text "-->" <+> prettyTCM cs1
+                              , text "--->" <+> prettyTCM cs1
                               ]
                           ]
                         patternViolation   -- TODO: will duplicate work (all arguments checked so far)
@@ -323,14 +336,21 @@ compareType cmp ty1@(El s1 a1) ty2@(El s2 a2) =
           [ hsep [ text "compareType", prettyTCM ty1, prettyTCM cmp, prettyTCM ty2 ]
           , hsep [ text "   sorts:", prettyTCM s1, text " and ", prettyTCM s2 ]
           ]
-	cs1 <- compareSort cmp s1 s2 `catchError` \err -> case err of
-                  TypeError _ _ -> typeError $ UnequalTypes cmp ty1 ty2
+	cs1 <- compareSort CmpEq s1 s2 `catchError` \err -> case errError err of
+                  TypeError _ _ -> do
+                    -- This error will probably be more informative
+                    compareTerm cmp (sort s1) a1 a2
+                    -- Throw the original error if the above doesn't
+                    -- give an error (for instance, due to pending
+                    -- constraints).
+                    throwError err
                   _             -> throwError err
 	cs2 <- compareTerm cmp (sort s1) a1 a2
-        unless (null $ cs1 ++ cs2) $
+        cs  <- solveConstraints $ cs1 ++ cs2
+        unless (null cs) $
           reportSDoc "tc.conv.type" 9 $
-            text "   --> " <+> prettyTCM (cs1 ++ cs2)
-	return $ cs1 ++ cs2
+            text "   --> " <+> prettyTCM cs
+	return cs
 
 leqType :: MonadTCM tcm => Type -> Type -> tcm Constraints
 leqType = compareType CmpLeq
@@ -341,7 +361,10 @@ leqType = compareType CmpLeq
 
 compareSort :: MonadTCM tcm => Comparison -> Sort -> Sort -> tcm Constraints
 compareSort CmpEq  = equalSort
-compareSort CmpLeq = equalSort -- TODO: change to leqSort when we have better constraint solving
+compareSort CmpLeq = leqSort
+  -- TODO: change to leqSort when we have better constraint solving
+  --       or not? might not be needed if we get universe polymorphism
+  --       leqSort is still used when checking datatype declarations though
 
 -- | Check that the first sort is less or equal to the second.
 leqSort :: MonadTCM tcm => Sort -> Sort -> tcm Constraints
@@ -349,9 +372,11 @@ leqSort s1 s2 =
   ifM typeInType (return []) $
     catchConstraint (SortCmp CmpLeq s1 s2) $
     do	(s1,s2) <- reduce (s1,s2)
--- 	do  d1 <- prettyTCM s1
--- 	    d2 <- prettyTCM s2
--- 	    debug $ "leqSort   " ++ show d1 ++ " <= " ++ show d2
+        reportSDoc "tc.conv.sort" 10 $
+          sep [ text "leqSort"
+              , nest 2 $ fsep [ prettyTCM s1 <+> text "=<"
+                              , prettyTCM s2 ]
+              ]
 	case (s1,s2) of
 
 	    (Prop    , Prop    )	     -> return []
@@ -359,60 +384,267 @@ leqSort s1 s2 =
 	    (Suc _   , Prop    )	     -> notLeq s1 s2
 
 	    (Prop    , Type _  )	     -> return []
-	    (Type n  , Type m  ) | n <= m    -> return []
-				 | otherwise -> notLeq s1 s2
-	    (Suc s   , Type n  ) | 1 <= n    -> leqSort s (Type $ n - 1)
-				 | otherwise -> notLeq s1 s2
+	    (Type (Lit (LitLevel _ n)), Type (Lit (LitLevel _ m)))
+              | n <= m    -> return []
+	      | otherwise -> notLeq s1 s2
+            (Type a, Type b) -> leqLevel a b
+
+	    (Suc s   , Type (Lit (LitLevel _ n)))
+              | 1 <= n    -> leqSort s (mkType $ n - 1)
+	    (Suc s   , Type b  ) -> notLeq s1 s2  -- TODO
+            (Suc s1  , Suc s2  ) -> leqSort s1 s2 -- TODO: not what we want for Prop(?)
 	    (_	     , Suc _   )	     -> equalSort s1 s2
 
 	    (Lub a b , _       )	     -> liftM2 (++) (leqSort a s2) (leqSort b s2)
 	    (_	     , Lub _ _ )	     -> equalSort s1 s2
 
-	    (MetaS x , MetaS y ) | x == y    -> return []
-	    (MetaS x , _       )	     -> equalSort s1 s2
-	    (_	     , MetaS x )	     -> equalSort s1 s2
+	    (MetaS{} , _       )	     -> equalSort s1 s2
+	    (_	     , MetaS{} )	     -> equalSort s1 s2
+            (Inf     , Inf     )             -> return []
+            (Inf     , _       )             -> notLeq s1 s2
+            (_       , Inf     )             -> return []
+            (DLub{}  , _       )             -> equalSort s1 s2
+            (_       , DLub{}  )             -> equalSort s1 s2
     where
 	notLeq s1 s2 = typeError $ NotLeqSort s1 s2
 
+leqLevel :: MonadTCM tcm => Term -> Term -> tcm Constraints
+leqLevel a b = liftTCM $ do
+  reportSDoc "tc.conv.nat" 10 $
+    text "compareLevel" <+>
+      sep [ prettyTCM a <+> text "=<"
+          , prettyTCM b ]
+  n <- levelView a
+  m <- levelView b
+  leqView n m
+  where
+    leqView n@(Max as) m@(Max bs) = do
+      reportSDoc "tc.conv.nat" 10 $
+        text "compareLevelView" <+>
+          sep [ text (show n) <+> text "=<"
+              , text (show m) ]
+      case (as, bs) of
+        (_, [])  -> concat <$> sequence [ leqPlusView a (ClosedLevel 0) | a <- as ]
+        (_, [b]) -> concat <$> sequence [ leqPlusView a b | a <- as ]
+        _        -> do
+          -- Each a needs to be less than at least one of the b's
+          sequence [ choice [ leqPlusView a b | b <- bs ] | a <- as ]
+          return []
+    leqPlusView n m = do
+      reportSDoc "tc.conv.nat" 10 $
+        text "comparePlusView" <+>
+          sep [ text (show n) <+> text "=<"
+              , text (show m) ]
+      case (n, m) of
+        -- Both closed
+        (ClosedLevel n, ClosedLevel m)
+          | n <= m -> ok
+
+        -- Both neutral
+        (Plus n (NeutralLevel a), Plus m (NeutralLevel b))
+          | n <= m -> do
+            lvl <- primLevel
+            equalTerm (El (mkType 0) lvl) a b
+
+        -- closed ≤ any
+        (ClosedLevel n, Plus m _)
+          | n <= m -> ok
+
+        -- Any blocked or meta
+        (Plus _ BlockedLevel{}, _) -> postpone
+        (_, Plus _ BlockedLevel{}) -> postpone
+        (Plus _ MetaLevel{}, _)    -> postpone
+        (_, Plus _ MetaLevel{})    -> postpone
+
+        _ -> notok
+    ok       = return []
+    notok    = typeError $ NotLeqSort (Type a) (Type b)
+    postpone = patternViolation
+
+    choice []     = patternViolation
+    choice (m:ms) = noConstraints m `catchError` \_ -> choice ms
+--       case errError e of
+--         PatternErr{} -> choice ms
+--         _            -> throwError e
+
+equalLevel :: MonadTCM tcm => Term -> Term -> tcm Constraints
+equalLevel a b = do
+  Max as <- levelView a
+  Max bs <- levelView b
+  a <- unLevelView (Max as)
+  b <- unLevelView (Max bs)
+  liftTCM $ check a b as bs
+  where
+    check a b as bs = do
+      let getLvl = El (mkType 0) <$> primLevel
+      reportSDoc "tc.conv.nat" 20 $
+        sep [ text "equalLevel"
+            , nest 2 $ sep [ prettyTCM a <+> text "=="
+                           , prettyTCM b
+                           ]
+            , nest 2 $ sep [ text (show (Max as)) <+> text "=="
+                           , text (show (Max bs))
+                           ]
+            ]
+      let a === b   = do
+            lvl <- getLvl
+            equalAtom lvl a b
+          as =!= bs = do a <- unLevelView (Max as)
+                         b <- unLevelView (Max bs)
+                         a === b
+      case (as, bs) of
+        _ | List.sort as == List.sort bs -> ok
+          | any isBlocked (as ++ bs) -> do
+              lvl <- getLvl
+              liftTCM $ useInjectivity CmpEq lvl a b
+
+        -- 0 == any
+        ([], _) -> wrap $ concat <$> sequence [ as =!= [b] | b <- bs ]
+        (_, []) -> wrap $ concat <$> sequence [ [a] =!= bs | a <- as ]
+
+        -- closed == closed
+        ([ClosedLevel n], [ClosedLevel m])
+          | n == m    -> ok
+          | otherwise -> notok
+
+        -- closed == neutral
+        ([ClosedLevel{}], _) | any isNeutral bs -> notok
+        (_, [ClosedLevel{}]) | any isNeutral as -> notok
+
+        -- meta == any
+        ([Plus n (MetaLevel x as)], _)
+          | any (isThisMeta x) bs -> postpone
+          | otherwise             -> do
+            bs' <- mapM (subtr n) bs
+            lvl <- getLvl
+            assignV lvl x as =<< unLevelView (Max bs')
+        (_, [Plus n (MetaLevel x bs)])
+          | any (isThisMeta x) as -> postpone
+          | otherwise             -> do
+            as' <- mapM (subtr n) as
+            lvl <- getLvl
+            assignV lvl x bs =<< unLevelView (Max as')
+
+        -- any other metas
+        _ | any isMeta (as ++ bs) -> postpone
+
+        -- neutral/closed == neutral/closed
+        _ | all isNeutralOrClosed (as ++ bs) -> as =!= bs
+
+        -- more cases?
+        _ -> postpone
+
+      where
+        ok       = return []
+        notok    = typeError $ UnequalSorts (Type a) (Type b)
+        postpone = patternViolation
+
+        -- Make sure to give a sensible error message
+        wrap m = m `catchError` \err ->
+          case errError err of
+            TypeError{} -> notok
+            _           -> throwError err
+
+        subtr n (ClosedLevel m)
+          | m >= n    = return $ ClosedLevel (m - n)
+          | otherwise = notok
+        subtr n (Plus m a)
+          | m >= n    = return $ Plus (m - n) a
+        subtr _ (Plus _ BlockedLevel{}) = postpone
+        subtr _ (Plus _ MetaLevel{})    = postpone
+        subtr _ (Plus _ NeutralLevel{}) = notok
+
+        isNeutral (Plus _ NeutralLevel{}) = True
+        isNeutral _                       = False
+
+        isClosed ClosedLevel{} = True
+        isClosed _             = False
+
+        isNeutralOrClosed l = isClosed l || isNeutral l
+
+        isBlocked (Plus _ BlockedLevel{}) = True
+        isBlocked _                     = False
+
+        isMeta (Plus _ MetaLevel{}) = True
+        isMeta _                  = False
+
+        isThisMeta x (Plus _ (MetaLevel y _)) = x == y
+        isThisMeta _ _                      = False
+
+
 -- | Check that the first sort equal to the second.
 equalSort :: MonadTCM tcm => Sort -> Sort -> tcm Constraints
 equalSort s1 s2 =
   ifM typeInType (return []) $
-    catchConstraint (SortCmp CmpEq s1 s2) $
-    do	(s1,s2) <- reduce (s1,s2)
--- 	do  d1 <- prettyTCM s1
--- 	    d2 <- prettyTCM s2
--- 	    debug $ "equalSort " ++ show d1 ++ " == " ++ show d2
+    catchConstraint (SortCmp CmpEq s1 s2) $ do
+        (s1,s2) <- reduce (s1,s2)
+        reportSDoc "tc.conv.sort" 10 $
+          sep [ text "equalSort"
+              , nest 2 $ fsep [ prettyTCM s1 <+> text "=="
+                              , prettyTCM s2 ]
+              ]
 	case (s1,s2) of
 
-	    (MetaS x , MetaS y ) | x == y    -> return []
-				 | otherwise -> do
+	    (MetaS x us , MetaS y vs) -> do
+              s1 <- normalise s1
+              s2 <- normalise s2
+              if s1 == s2 then return []
+			  else do
 		[p1, p2] <- mapM getMetaPriority [x, y]
-		if p1 >= p2 then assignS x s2
-			    else assignS y s1
-	    (MetaS x , _       )	     -> assignS x s2
-	    (_	     , MetaS x )	     -> equalSort s2 s1
+		if p1 >= p2 then assignS x us s2
+			    else assignS y vs s1
+	    (MetaS x vs, _       ) -> assignS x vs s2
+	    (_	     , MetaS{} ) -> equalSort s2 s1
 
-	    (Prop    , Prop    )	     -> return []
-	    (Type _  , Prop    )	     -> notEq s1 s2
-	    (Prop    , Type _  )	     -> notEq s1 s2
+	    (Prop    , Prop    ) -> return []
+	    (Type _  , Prop    ) -> notEq s1 s2
+	    (Prop    , Type _  ) -> notEq s1 s2
 
-	    (Type n  , Type m  ) | n == m    -> return []
-				 | otherwise -> notEq s1 s2
-	    (Suc s   , Prop    )	     -> notEq s1 s2
-	    (Suc s   , Type 0  )	     -> notEq s1 s2
-	    (Suc s   , Type 1  )	     -> buildConstraint (SortCmp CmpEq s1 s2)
-	    (Suc s   , Type n  )	     -> equalSort s (Type $ n - 1)
-	    (Prop    , Suc s   )	     -> notEq s1 s2
-	    (Type 0  , Suc s   )	     -> notEq s1 s2
-	    (Type 1  , Suc s   )	     -> buildConstraint (SortCmp CmpEq s1 s2)
-	    (Type n  , Suc s   )	     -> equalSort (Type $ n - 1) s
-	    (_	     , Suc _   )	     -> buildConstraint (SortCmp CmpEq s1 s2)
-	    (Suc _   , _       )	     -> buildConstraint (SortCmp CmpEq s1 s2)
+            (Type (Lit n), Type (Lit m))
+              | n == m           -> return []
+              | otherwise        -> notEq s1 s2
+            (Type a  , Type b  ) -> equalLevel a b
+	    (Suc s   , Prop    ) -> notEq s1 s2
+	    (Suc s   , Type (Lit (LitLevel _ 0))) -> notEq s1 s2
+	    (Suc s   , Type (Lit (LitLevel _ 1))) -> buildConstraint (SortCmp CmpEq s1 s2)
+	    (Suc s   , Type (Lit (LitLevel _ n))) -> equalSort s (mkType $ n - 1)
+	    (Prop    , Suc s   ) -> notEq s1 s2
+	    (Type (Lit (LitLevel _ 0))  , Suc s   ) -> notEq s1 s2
+	    (Type (Lit (LitLevel _ 1))  , Suc s   ) -> buildConstraint (SortCmp CmpEq s1 s2)
+	    (Type (Lit (LitLevel _ n))  , Suc s   ) -> equalSort (mkType $ n - 1) s
+            (Suc s1  , Suc s2  ) -> equalSort s1 s2 -- TODO: not what we want for Prop(?)
+	    (_	     , Suc _   ) -> buildConstraint (SortCmp CmpEq s1 s2)
+	    (Suc _   , _       ) -> buildConstraint (SortCmp CmpEq s1 s2)
 
-	    (Lub _ _ , _       )	     -> buildConstraint (SortCmp CmpEq s1 s2)
-	    (_	     , Lub _ _ )	     -> buildConstraint (SortCmp CmpEq s1 s2)
+            (Inf     , Inf     )             -> return []
+            (Inf     , _       )             -> notEq s1 s2
+            (_       , Inf     )             -> notEq s1 s2
+            (s0@(Type (Lit (LitLevel _ 0))), Lub s1 s2) -> do
+              cs1 <- equalSort s0 s1
+              cs2 <- equalSort s0 s2
+              return $ cs1 ++ cs2
+            (Lub s1 s2, s0@(Type (Lit (LitLevel _ 0)))) -> do
+              cs1 <- equalSort s1 s0
+              cs2 <- equalSort s2 s0
+              return $ cs1 ++ cs2
+            (Lub{}   , Lub{}   ) -> do
+              (s1, s2) <- normalise (s1, s2)
+              if s1 == s2
+                then return []
+                else buildConstraint (SortCmp CmpEq s1 s2)
+	    (Lub _ _ , _       ) -> buildConstraint (SortCmp CmpEq s1 s2)
+	    (_	     , Lub _ _ ) -> buildConstraint (SortCmp CmpEq s1 s2)
 
+            (DLub s1 s2, s0@(Type (Lit (LitLevel _ 0)))) -> do
+              cs1 <- equalSort s1 s0
+              cs2 <- underAbstraction_ s2 $ \s2 -> equalSort s2 s0
+              return $ cs1 ++ cs2
+            (s0@(Type (Lit (LitLevel _ 0))), DLub s1 s2) -> do
+              cs1 <- equalSort s0 s1
+              cs2 <- underAbstraction_ s2 $ \s2 -> equalSort s0 s2
+              return $ cs1 ++ cs2
+            (DLub{}  , _       )             -> buildConstraint (SortCmp CmpEq s1 s2)
+            (_       , DLub{}  )             -> buildConstraint (SortCmp CmpEq s1 s2)
     where
 	notEq s1 s2 = typeError $ UnequalSorts s1 s2
 
