@@ -2,6 +2,7 @@
 module Agda.TypeChecking.Errors
     ( prettyError
     , PrettyTCM(..)
+    , tcErrString
     ) where
 
 import Control.Applicative ( (<$>) )
@@ -26,6 +27,7 @@ import Agda.Syntax.Translation.AbstractToConcrete
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Pretty
 
+import Agda.Utils.FileName
 import Agda.Utils.Monad
 import Agda.Utils.Trace
 import Agda.Utils.Size
@@ -69,11 +71,12 @@ nameWithBinding q =
     r = nameBindingSite $ qnameName q
 
 tcErrString :: TCErr -> String
-tcErrString err = show (getRange err) ++ " " ++ case err of
-    TypeError _ cl -> errorString $ clValue cl
-    Exception r s  -> show r ++ " " ++ s
-    PatternErr _   -> "PatternErr"
-    AbortAssign _  -> "AbortAssign"
+tcErrString err = show (getRange err) ++ " " ++ case errError err of
+    TypeError _ cl  -> errorString $ clValue cl
+    Exception r s   -> show r ++ " " ++ s
+    IOException r e -> show r ++ " " ++ show e
+    PatternErr _    -> "PatternErr"
+    AbortAssign _   -> "AbortAssign"
 
 errorString :: TypeError -> String
 errorString err = case err of
@@ -81,6 +84,7 @@ errorString err = case err of
     AmbiguousName _ _			       -> "AmbiguousName"
     AmbiguousParseForApplication _ _	       -> "AmbiguousParseForApplication"
     AmbiguousParseForLHS _ _		       -> "AmbiguousParseForLHS"
+    AmbiguousTopLevelModuleName {}	       -> "AmbiguousTopLevelModuleName"
     BothWithAndRHS                             -> "BothWithAndRHS"
     BuiltinInParameterisedModule _	       -> "BuiltinInParameterisedModule"
     BuiltinMustBeConstructor _ _               -> "BuiltinMustBeConstructor"
@@ -113,8 +117,9 @@ errorString err = case err of
     MetaCannotDependOn _ _ _		       -> "MetaCannotDependOn"
     MetaOccursInItself _		       -> "MetaOccursInItself"
     ModuleArityMismatch _ _ _                  -> "ModuleArityMismatch"
+    ModuleDefinedInOtherFile {}                -> "ModuleDefinedInOtherFile"
     ModuleDoesntExport _ _		       -> "ModuleDoesntExport"
-    ModuleNameDoesntMatchFileName _	       -> "ModuleNameDoesntMatchFileName"
+    ModuleNameDoesntMatchFileName {}	       -> "ModuleNameDoesntMatchFileName"
     NoBindingForBuiltin _		       -> "NoBindingForBuiltin"
     NoParseForApplication _		       -> "NoParseForApplication"
     NoParseForLHS _			       -> "NoParseForLHS"
@@ -134,6 +139,7 @@ errorString err = case err of
     NotLeqSort _ _			       -> "NotLeqSort"
     NotStrictlyPositive _ _		       -> "NotStrictlyPositive"
     NothingAppliedToHiddenArg _		       -> "NothingAppliedToHiddenArg"
+    OverlappingProjects {}                     -> "OverlappingProjects"
     PatternShadowsConstructor {}               -> "PatternShadowsConstructor"
     PropMustBeSingleton			       -> "PropMustBeSingleton"
     RepeatedVariablesInPattern _	       -> "RepeatedVariablesInPattern"
@@ -167,16 +173,17 @@ errorString err = case err of
     WrongNumberOfConstructorArguments _ _ _    -> "WrongNumberOfConstructorArguments"
 
 instance PrettyTCM TCErr where
-    prettyTCM err = case err of
+    prettyTCM err = case errError err of
 	TypeError s e -> do
 	    s0 <- get
 	    put s
 	    d <- sayWhen (clTrace e) $ prettyTCM e
 	    put s0
 	    return d
-	Exception r s -> sayWhere r $ fwords s
-	PatternErr _  -> sayWhere err $ panic "uncaught pattern violation"
-	AbortAssign _ -> sayWhere err $ panic "uncaught aborted assignment"
+	Exception r s   -> sayWhere r $ fwords s
+	IOException r e -> sayWhere r $ fwords $ show e
+	PatternErr _    -> sayWhere err $ panic "uncaught pattern violation"
+	AbortAssign _   -> sayWhere err $ panic "uncaught aborted assignment"
 
 instance PrettyTCM TypeError where
     prettyTCM err = do
@@ -348,23 +355,35 @@ instance PrettyTCM TypeError where
 		$$ nest 2 (vcat $ map prettyTCM cs)
 	    CyclicModuleDependency ms ->
 		fsep (pwords "cyclic module dependency:")
-		$$ nest 2 (vcat $ map (text . show) ms)
+		$$ nest 2 (vcat $ map pretty ms)
 	    FileNotFound x files ->
-		fsep ( pwords "Failed to find source of module" ++ [text $ show x] ++
+		fsep ( pwords "Failed to find source of module" ++ [pretty x] ++
 		       pwords "in any of the following locations:"
-		     ) $$ nest 2 (vcat $ map text files)
+		     ) $$ nest 2 (vcat $ map (text . filePath) files)
+            OverlappingProjects f m1 m2 ->
+	        fsep ( pwords "The file" ++ [text (filePath f)] ++
+                       pwords "can be accessed via several project roots. Both" ++
+                       [pretty m1] ++ pwords "and" ++ [pretty m2] ++
+                       pwords "point to this file."
+                     )
+	    AmbiguousTopLevelModuleName x files ->
+		fsep ( pwords "Ambiguous module name. The module name" ++
+                       [pretty x] ++
+		       pwords "could refer to any of the following files:"
+		     ) $$ nest 2 (vcat $ map (text . filePath) files)
 	    ClashingFileNamesFor x files ->
 		fsep ( pwords "Multiple possible sources for module" ++ [text $ show x] ++
 		       pwords "found:"
-		     ) $$ nest 2 (vcat $ map text files)
-	    ModuleNameDoesntMatchFileName given -> fsep $
-              pwords "The name of the top level module does not match the file name. The module" ++
-              [ text (show given) ] ++ pwords "should be defined in either" ++
-              [ text ("<top-level>" </> mod ".agda")
-              , text "or"
-              , text ("<top-level>" </> mod ".lagda") <> text "."
-              ]
-              where mod = C.moduleNameToFileName (mnameToConcrete given)
+		     ) $$ nest 2 (vcat $ map (text . filePath) files)
+            ModuleDefinedInOtherFile mod file file' -> fsep $
+              pwords "You tried to load" ++ [text (filePath file)] ++
+              pwords "which defines the module" ++ [pretty mod <> text "."] ++
+              pwords "However, according to the include path this module should" ++
+              pwords "be defined in" ++ [text (filePath file') <> text "."]
+	    ModuleNameDoesntMatchFileName given files ->
+              fsep (pwords "The name of the top level module does not match the file name. The module" ++
+                   [ pretty given ] ++ pwords "should be defined in one of the following files:")
+	      $$ nest 2 (vcat $ map (text . filePath) files)
             BothWithAndRHS -> fsep $
               pwords "Unexpected right hand side"
 	    NotInScope xs ->
@@ -584,8 +603,8 @@ instance PrettyTCM Call where
 		[ C.Data r ind d (map bind bs) (C.Underscore noRange Nothing)
 		    $ map simpleDecl cs
 		]
-	      D.RecDef r fx p a d bs cs ->
-		[ C.Record r d (map bind bs) (C.Underscore noRange Nothing)
+	      D.RecDef r fx p a d c bs cs ->
+		[ C.Record r d (name <$> c) (map bind bs) (C.Underscore noRange Nothing)
 		    $ map simpleDecl cs
 		]
 	      where
@@ -593,11 +612,13 @@ instance PrettyTCM Call where
 		bind (C.DomainFull b) = b
 		bind (C.DomainFree h x) = C.TypedBindings r h [C.TBind r [x] (C.Underscore r Nothing)]
 		  where r = getRange x
-		-- bind _		      = __IMPOSSIBLE__
+
+                name (D.Axiom _ _ _ _ n _) = n
+                name _                     = __IMPOSSIBLE__
 
 	    simpleDecl d = case d of
 		D.Axiom _ _ _ _ x e		       -> C.TypeSig x e
-		D.NiceField _ _ _ _ x e		       -> C.Field x e
+		D.NiceField _ _ _ _ h x e	       -> C.Field h x e
 		D.PrimitiveFunction r _ _ _ x e	       -> C.Primitive r [C.TypeSig x e]
 		D.NiceDef r ds _ _		       -> C.Mutual r ds
 		D.NiceModule r _ _ x tel _	       -> C.Module r x tel []
