@@ -1,8 +1,8 @@
 
 module Agda.TypeChecking.Monad.Debug where
 
-import qualified System.IO.UTF8 as UTF8
+import qualified Agda.Utils.IO.Locale as LocIO
 import Control.Monad.Trans ( MonadIO(liftIO) )
 
 debug :: MonadIO m => String -> m ()
-debug s = liftIO $ UTF8.putStrLn s
\ No newline at end of file
+debug s = liftIO $ LocIO.putStrLn s
\ No newline at end of file
