@@ -5,9 +5,11 @@ module Agda.TypeChecking.Reduce where
 import Prelude hiding (mapM)
 import Control.Monad.State hiding (mapM)
 import Control.Monad.Reader hiding (mapM)
+import Control.Monad.Error hiding (mapM)
 import Control.Applicative
 import Data.List as List hiding (sort)
 import Data.Map as Map
+import qualified Data.Set as Set
 import Data.Generics
 import Data.Traversable
 
@@ -21,6 +23,7 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Free
 
 import {-# SOURCE #-} Agda.TypeChecking.Patterns.Match
 
@@ -29,6 +32,20 @@ import Agda.Utils.Monad
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
+traceFun :: MonadTCM tcm => String -> tcm a -> tcm a
+traceFun s m = do
+  reportSLn "tc.inst" 100 $ "[ " ++ s
+  x <- m
+  reportSLn "tc.inst" 100 $ "]"
+  return x
+
+traceFun' :: (Show a, MonadTCM tcm) => String -> tcm a -> tcm a
+traceFun' s m = do
+  reportSLn "tc.inst" 100 $ "[ " ++ s
+  x <- m
+  reportSLn "tc.inst" 100 $ "  result = " ++ show x ++ "\n]"
+  return x
+
 -- | Instantiate something.
 --   Results in an open meta variable or a non meta.
 --   Doesn't do any reduction, and preserves blocking tags (when blocking meta
@@ -37,15 +54,16 @@ class Instantiate t where
     instantiate :: MonadTCM tcm => t -> tcm t
 
 instance Instantiate Term where
-    instantiate t@(MetaV x args) =
-	do  mi <- mvInstantiation <$> lookupMeta x
-	    case mi of
-		InstV a                        -> instantiate $ a `apply` args
-		Open                           -> return t
-		BlockedConst _                 -> return t
-                PostponedTypeCheckingProblem _ -> return t
-		InstS _                        -> __IMPOSSIBLE__
-    instantiate t = return t
+  instantiate t@(MetaV x args) = do
+    mi <- mvInstantiation <$> lookupMeta x
+    case mi of
+      InstV a                        -> instantiate $ a `apply` args
+      Open                           -> return t
+      BlockedConst _                 -> return t
+      PostponedTypeCheckingProblem _ -> return t
+      InstS _                        -> __IMPOSSIBLE__
+  instantiate (Sort s) = Sort <$> instantiate s
+  instantiate t = return t
 
 instance Instantiate a => Instantiate (Blocked a) where
   instantiate v@NotBlocked{} = return v
@@ -59,22 +77,26 @@ instance Instantiate a => Instantiate (Blocked a) where
       InstS _                        -> __IMPOSSIBLE__
 
 instance Instantiate Type where
-    instantiate (El s t) = El s <$> instantiate t
+    instantiate (El s t) = El <$> instantiate s <*> instantiate t
 
 instance Instantiate Sort where
     instantiate s = case s of
-	MetaS x -> do
+	MetaS x as -> do
 	    mi <- mvInstantiation <$> lookupMeta x
 	    case mi of
-		InstS s'                       -> instantiate s'
+		InstS s'                       -> do
+                  v <- instantiate (s' `apply` as)
+                  case v of
+                    Sort s -> return s
+                    _      -> __IMPOSSIBLE__
 		Open                           -> return s
 		InstV{}                        -> __IMPOSSIBLE__
 		BlockedConst{}                 -> __IMPOSSIBLE__
                 PostponedTypeCheckingProblem{} -> __IMPOSSIBLE__
-	Type _	  -> return s
-	Prop	  -> return s
-	Suc s	  -> sSuc <$> instantiate s
-	Lub s1 s2 -> sLub <$> instantiate s1 <*> instantiate s2
+        _ -> return s
+
+instance Instantiate t => Instantiate (Abs t) where
+  instantiate = traverse instantiate
 
 instance Instantiate t => Instantiate (Arg t) where
     instantiate = traverse instantiate
@@ -101,12 +123,12 @@ instance Instantiate Constraint where
     instantiate (ValueCmp cmp t u v) =
 	do  (t,u,v) <- instantiate (t,u,v)
 	    return $ ValueCmp cmp t u v
-    instantiate (TypeCmp cmp a b) = uncurry (TypeCmp cmp) <$> instantiate (a,b)
-    instantiate (TelCmp cmp a b)  = uncurry (TelCmp cmp)  <$> instantiate (a,b)
-    instantiate (SortCmp cmp a b) = uncurry (SortCmp cmp) <$> instantiate (a,b)
-    instantiate (Guarded c cs)    = uncurry Guarded <$> instantiate (c,cs)
-    instantiate (UnBlock m)       = return $ UnBlock m
-    instantiate (IsEmpty t)       = IsEmpty <$> instantiate t
+    instantiate (TypeCmp cmp a b)    = uncurry (TypeCmp cmp) <$> instantiate (a,b)
+    instantiate (TelCmp cmp a b)     = uncurry (TelCmp cmp)  <$> instantiate (a,b)
+    instantiate (SortCmp cmp a b)    = uncurry (SortCmp cmp) <$> instantiate (a,b)
+    instantiate (Guarded c cs)       = uncurry Guarded <$> instantiate (c,cs)
+    instantiate (UnBlock m)          = return $ UnBlock m
+    instantiate (IsEmpty t)          = IsEmpty <$> instantiate t
 
 instance (Ord k, Instantiate e) => Instantiate (Map k e) where
     instantiate = traverse instantiate
@@ -131,15 +153,39 @@ instance Reduce Type where
       return (El s <$> t)
 
 instance Reduce Sort where
-    reduce s =
-	{-# SCC "reduce<Sort>" #-}
-	do  s <- instantiate s
-	    case s of
-		Suc s'	  -> sSuc <$> reduce s'
-		Lub s1 s2 -> sLub <$> reduce s1 <*> reduce s2
-		Prop	  -> return s
-		Type _	  -> return s
-		MetaS _   -> return s
+    reduce s = {-# SCC "reduce<Sort>" #-}
+      ifM (not <$> hasUniversePolymorphism) (red sSuc sLub s)
+      $ liftTCM $ do
+        suc  <- do sc <- primLevelSuc
+                   return $ \x -> case x of
+                                    Type n -> Type (sc `apply` [Arg NotHidden n])
+                                    _      -> sSuc x
+          `catchError` \_ -> return sSuc
+        lub <- do mx <- primLevelMax
+                  return $ \x y -> case (x, y) of
+                    (Type n, Type m) -> Type (mx `apply` List.map (Arg NotHidden) [n, m])
+                    _                -> sLub x y
+          `catchError` \_ -> return sLub
+        -- need to instantiate all meta to know if we need a DLub or a Lub
+        red suc lub =<< instantiateFull s
+      where
+        red suc lub s = do
+          s <- instantiate s
+          case s of
+            Suc s'     -> suc <$> reduce s'
+            Lub s1 s2  -> lub <$> reduce s1 <*> reduce s2
+            DLub s1 s2 -> do
+              s <- dLub <$> reduce s1 <*> reduce s2
+              case s of
+                DLub{}  -> return s
+                _       -> reduce s   -- TODO: not so nice
+            Prop       -> return s
+            Type s'    -> Type <$> reduce s'
+            MetaS m as -> MetaS m <$> reduce as
+            Inf        -> return Inf
+
+instance Reduce t => Reduce (Abs t) where
+  reduce b = Abs (absName b) <$> underAbstraction_ b reduce
 
 -- Lists are never blocked
 instance Reduce t => Reduce [t] where
@@ -176,19 +222,24 @@ instance Reduce Term where
 		Lam _ _    -> return $ notBlocked v
 	where
 	    reduceNat v@(Con c []) = do
-		mz <- getBuiltin' builtinZero
-		case mz of
-		    Just (Con z []) | c == z -> return $ Lit $ LitInt (getRange c) 0
-		    _			     -> return v
+		mz  <- getBuiltin' builtinZero
+                mlz <- getBuiltin' builtinLevelZero
+		case v of
+                  _ | Just v == mz  -> return $ Lit $ LitInt (getRange c) 0
+                    | Just v == mlz -> return $ Lit $ LitLevel (getRange c) 0
+		  _		    -> return v
 	    reduceNat v@(Con c [Arg NotHidden w]) = do
-		ms <- getBuiltin' builtinSuc
-		case ms of
-		    Just (Con s []) | c == s -> do
-			w <- reduce w
-			case w of
-			    Lit (LitInt r n) -> return $ Lit $ LitInt (fuseRange c r) $ n + 1
-			    _		     -> return $ Con c [Arg NotHidden w]
-		    _	-> return v
+		ms  <- getBuiltin' builtinSuc
+                mls <- getBuiltin' builtinLevelSuc
+		case v of
+                  _ | Just (Con c []) == ms ||
+                      Just (Con c []) == mls -> inc <$> reduce w
+		  _	                     -> return v
+                  where
+                    inc w = case w of
+                      Lit (LitInt r n) -> Lit (LitInt (fuseRange c r) $ n + 1)
+                      Lit (LitLevel r n) -> Lit (LitLevel (fuseRange c r) $ n + 1)
+                      _                  -> Con c [Arg NotHidden w]
 	    reduceNat v = return v
 
 -- | If the first argument is 'True', then a single delayed clause may
@@ -204,12 +255,12 @@ unfoldDefinition unfoldDelayed keepGoing v0 f args =
               return $ notBlocked $ Con (c `withRangeOf` f) args
             Primitive ConcreteDef x cls -> do
                 pf <- getPrimitive x
-                reducePrimitive x v0 f args pf (defDelayed info) cls
+                reducePrimitive x v0 f args pf (defDelayed info) (maybe [] id cls)
             _  -> reduceNormal v0 f args (defDelayed info) (defClauses info)
   where
     reducePrimitive x v0 f args pf delayed cls
         | n < ar    = return $ notBlocked $ v0 `apply` args -- not fully applied
-        | otherwise = do
+        | otherwise = {-# SCC "reducePrimitive" #-} do
             let (args1,args2) = genericSplitAt ar args
             r <- def args1
             case r of
@@ -221,7 +272,7 @@ unfoldDefinition unfoldDelayed keepGoing v0 f args =
             ar  = primFunArity pf
             def = primFunImplementation pf
 
-    reduceNormal v0 f args delayed def = do
+    reduceNormal v0 f args delayed def = {-# SCC "reduceNormal" #-} do
         case (delayed, def) of
             (Delayed, _) | not unfoldDelayed -> defaultResult
             (_, []) -> defaultResult -- no definition for head
@@ -238,7 +289,7 @@ unfoldDefinition unfoldDelayed keepGoing v0 f args =
     -- Apply a defined function to it's arguments.
     --   The original term is the first argument applied to the third.
     appDef :: MonadTCM tcm => Term -> [Clause] -> Args -> tcm (Reduced (Blocked Term) Term)
-    appDef v cls args = goCls cls args where
+    appDef v cls args = {-# SCC "appDef" #-} goCls cls args where
 
         goCls :: MonadTCM tcm => [Clause] -> Args -> tcm (Reduced (Blocked Term) Term)
         goCls [] args = typeError $ IncompletePatternMatching v args
@@ -262,7 +313,7 @@ unfoldDefinition unfoldDelayed keepGoing v0 f args =
         hasBody (NoBind b)	 = hasBody b
 
         app []		 (Body v')	     = v'
-        app (arg : args) (Bind (Abs _ body)) = app args $ subst arg body -- CBN
+        app (arg : args) (Bind (Abs _ body)) = {-# SCC "instRHS" #-} app args $ subst arg body -- CBN
         app (_   : args) (NoBind body)	     = app args body
         app  _		  NoBody	     = __IMPOSSIBLE__
         app (_ : _)	 (Body _)	     = __IMPOSSIBLE__
@@ -282,12 +333,12 @@ instance Reduce Constraint where
     reduce (ValueCmp cmp t u v) =
 	do  (t,u,v) <- reduce (t,u,v)
 	    return $ ValueCmp cmp t u v
-    reduce (TypeCmp cmp a b) = uncurry (TypeCmp cmp) <$> reduce (a,b)
-    reduce (TelCmp  cmp a b) = uncurry (TelCmp cmp)  <$> reduce (a,b)
-    reduce (SortCmp cmp a b) = uncurry (SortCmp cmp) <$> reduce (a,b)
-    reduce (Guarded c cs)    = uncurry Guarded <$> reduce (c,cs)
-    reduce (UnBlock m)       = return $ UnBlock m
-    reduce (IsEmpty t)       = IsEmpty <$> reduce t
+    reduce (TypeCmp cmp a b)    = uncurry (TypeCmp cmp) <$> reduce (a,b)
+    reduce (TelCmp  cmp a b)    = uncurry (TelCmp cmp)  <$> reduce (a,b)
+    reduce (SortCmp cmp a b)    = uncurry (SortCmp cmp) <$> reduce (a,b)
+    reduce (Guarded c cs)       = uncurry Guarded <$> reduce (c,cs)
+    reduce (UnBlock m)          = return $ UnBlock m
+    reduce (IsEmpty t)          = IsEmpty <$> reduce t
 
 instance (Ord k, Reduce e) => Reduce (Map k e) where
     reduce = traverse reduce
@@ -300,7 +351,16 @@ class Normalise t where
     normalise :: MonadTCM tcm => t -> tcm t
 
 instance Normalise Sort where
-    normalise = reduce
+    normalise s = do
+      s <- reduce s
+      case s of
+        Suc s'     -> sSuc <$> normalise s'
+        Lub s1 s2  -> sLub <$> normalise s1 <*> normalise s2
+        DLub s1 s2 -> dLub <$> normalise s1 <*> normalise s2
+        Prop       -> return s
+        Type s'    -> Type <$> normalise s'
+        MetaS m as -> MetaS m <$> normalise as
+        Inf        -> return Inf
 
 instance Normalise Type where
     normalise (El s t) = El <$> normalise s <*> normalise t
@@ -355,12 +415,12 @@ instance Normalise Constraint where
     normalise (ValueCmp cmp t u v) =
 	do  (t,u,v) <- normalise (t,u,v)
 	    return $ ValueCmp cmp t u v
-    normalise (TypeCmp cmp a b) = uncurry (TypeCmp cmp) <$> normalise (a,b)
-    normalise (TelCmp cmp a b)  = uncurry (TelCmp cmp)  <$> normalise (a,b)
-    normalise (SortCmp cmp a b) = uncurry (SortCmp cmp) <$> normalise (a,b)
-    normalise (Guarded c cs)    = uncurry Guarded <$> normalise (c,cs)
-    normalise (UnBlock m)       = return $ UnBlock m
-    normalise (IsEmpty t)       = IsEmpty <$> normalise t
+    normalise (TypeCmp cmp a b)    = uncurry (TypeCmp cmp) <$> normalise (a,b)
+    normalise (TelCmp cmp a b)     = uncurry (TelCmp cmp)  <$> normalise (a,b)
+    normalise (SortCmp cmp a b)    = uncurry (SortCmp cmp) <$> normalise (a,b)
+    normalise (Guarded c cs)       = uncurry Guarded <$> normalise (c,cs)
+    normalise (UnBlock m)          = return $ UnBlock m
+    normalise (IsEmpty t)          = IsEmpty <$> normalise t
 
 instance Normalise Pattern where
   normalise p = case p of
@@ -393,28 +453,31 @@ instance InstantiateFull Sort where
     instantiateFull s = do
 	s <- instantiate s
 	case s of
-	    MetaS x   -> return $ MetaS x
-	    Type _    -> return s
-	    Prop      -> return s
-	    Suc s     -> sSuc <$> instantiateFull s
-	    Lub s1 s2 -> sLub <$> instantiateFull s1 <*> instantiateFull s2
+	    MetaS x as -> MetaS x <$> instantiateFull as
+	    Type n     -> Type <$> instantiateFull n
+	    Prop       -> return s
+	    Suc s      -> sSuc <$> instantiateFull s
+	    Lub s1 s2  -> sLub <$> instantiateFull s1 <*> instantiateFull s2
+	    DLub s1 s2 -> dLub <$> instantiateFull s1 <*> instantiateFull s2
+            Inf        -> return s
 
 instance InstantiateFull Type where
-    instantiateFull (El s t) = El <$> instantiateFull s <*> instantiateFull t
+    instantiateFull (El s t) =
+      El <$> instantiateFull s <*> instantiateFull t
 
 instance InstantiateFull Term where
-    instantiateFull v =
-	do  v <- instantiate v
-	    case v of
-		Var n vs   -> Var n <$> instantiateFull vs
-		Con c vs   -> Con c <$> instantiateFull vs
-		Def f vs   -> Def f <$> instantiateFull vs
-		MetaV x vs -> MetaV x <$> instantiateFull vs
-		Lit _	   -> return v
-		Lam h b    -> Lam h <$> instantiateFull b
-		Sort s	   -> Sort <$> instantiateFull s
-		Pi a b	   -> uncurry Pi <$> instantiateFull (a,b)
-		Fun a b    -> uncurry Fun <$> instantiateFull (a,b)
+    instantiateFull v = do
+      v <- instantiate v
+      case v of
+          Var n vs   -> Var n <$> instantiateFull vs
+          Con c vs   -> Con c <$> instantiateFull vs
+          Def f vs   -> Def f <$> instantiateFull vs
+          MetaV x vs -> MetaV x <$> instantiateFull vs
+          Lit _	   -> return v
+          Lam h b    -> Lam h <$> instantiateFull b
+          Sort s	   -> Sort <$> instantiateFull s
+          Pi a b	   -> uncurry Pi <$> instantiateFull (a,b)
+          Fun a b    -> uncurry Fun <$> instantiateFull (a,b)
 
 instance InstantiateFull ClauseBody where
     instantiateFull (Body   t) = Body   <$> instantiateFull t
@@ -445,15 +508,16 @@ instance InstantiateFull a => InstantiateFull (Closure a) where
 	return $ cl { clValue = x }
 
 instance InstantiateFull Constraint where
-    instantiateFull (ValueCmp cmp t u v) =
-	do  (t,u,v) <- instantiateFull (t,u,v)
-	    return $ ValueCmp cmp t u v
-    instantiateFull (TypeCmp cmp a b) = uncurry (TypeCmp cmp) <$> instantiateFull (a,b)
-    instantiateFull (TelCmp cmp a b)  = uncurry (TelCmp cmp)  <$> instantiateFull (a,b)
-    instantiateFull (SortCmp cmp a b) = uncurry (SortCmp cmp) <$> instantiateFull (a,b)
-    instantiateFull (Guarded c cs)    = uncurry Guarded <$> instantiateFull (c,cs)
-    instantiateFull (UnBlock m)       = return $ UnBlock m
-    instantiateFull (IsEmpty t)       = IsEmpty <$> instantiateFull t
+  instantiateFull c = case c of
+    ValueCmp cmp t u v -> do
+      (t,u,v) <- instantiateFull (t,u,v)
+      return $ ValueCmp cmp t u v
+    TypeCmp cmp a b    -> uncurry (TypeCmp cmp) <$> instantiateFull (a,b)
+    TelCmp cmp a b     -> uncurry (TelCmp cmp)  <$> instantiateFull (a,b)
+    SortCmp cmp a b    -> uncurry (SortCmp cmp) <$> instantiateFull (a,b)
+    Guarded c cs       -> uncurry Guarded <$> instantiateFull (c,cs)
+    UnBlock m          -> return $ UnBlock m
+    IsEmpty t          -> IsEmpty <$> instantiateFull t
 
 instance (Ord k, InstantiateFull e) => InstantiateFull (Map k e) where
     instantiateFull = traverse instantiateFull
@@ -502,11 +566,11 @@ instance InstantiateFull Defn where
 	s  <- instantiateFull s
 	cl <- instantiateFull cl
 	return $ d { dataSort = s, dataClause = cl }
-      Record{ recSort = s, recClause = cl, recTel = tel } -> do
-        s   <- instantiateFull s
+      Record{ recConType = t, recClause = cl, recTel = tel } -> do
+        t   <- instantiateFull t
         cl  <- instantiateFull cl
         tel <- instantiateFull tel
-        return $ d { recSort = s, recClause = cl, recTel = tel }
+        return $ d { recConType = t, recClause = cl, recTel = tel }
       Constructor{} -> return d
       Primitive{ primClauses = cs } -> do
         cs <- instantiateFull cs
@@ -524,8 +588,9 @@ instance InstantiateFull Clause where
        <*> instantiateFull b
 
 instance InstantiateFull Interface where
-    instantiateFull (Interface ms mod scope sig b hsImports highlighting) =
-	Interface ms mod scope
+    instantiateFull (Interface ms mod scope inside
+                               sig b hsImports highlighting) =
+	Interface ms mod scope inside
 	    <$> instantiateFull sig
 	    <*> instantiateFull b
             <*> return hsImports
