@@ -11,7 +11,7 @@ import Data.List hiding (sort)
 import Data.Traversable
 import Data.Set (Set)
 import qualified Data.Set as Set
-import qualified System.IO.UTF8 as UTF8
+import qualified Agda.Utils.IO.Locale as LocIO
 
 import Agda.Syntax.Common
 import Agda.Syntax.Position
@@ -21,8 +21,10 @@ import qualified Agda.Syntax.Info as Info
 import qualified Agda.Syntax.Abstract.Pretty as A
 import Agda.Syntax.Fixity
 import Agda.Syntax.Translation.InternalToAbstract
-
+import Agda.Syntax.Info
+import Agda.Syntax.Scope.Base (emptyScopeInfo)
 import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Monad.Builtin (primRefl, primEquality)
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
@@ -106,13 +108,24 @@ checkFunDef delayed i name cs =
         verboseS "tc.def.fun" 10 $ do
           dx <- prettyTCM name
           t' <- prettyTCM . defType =<< getConstInfo name
-          liftIO $ UTF8.putStrLn $ "added " ++ show dx ++ " : " ++ show t'
+          liftIO $ LocIO.putStrLn $ "added " ++ show dx ++ " : " ++ show t'
 
         -- Check pattern coverage
         checkCoverage name
     where
         npats = size . clausePats
 
+
+-- | Insert some patterns in the in with-clauses LHS of the given RHS
+insertPatterns :: [A.Pattern] -> A.RHS -> A.RHS
+insertPatterns pats (A.WithRHS aux es cs) = A.WithRHS aux es (map insertToClause cs)
+    where insertToClause (A.Clause (A.LHS i x aps ps) rhs ds) 
+--              = A.Clause (A.LHS i x (aps ++ map (Arg NotHidden . unnamed) pats) (ps)) (insertPatterns pats rhs) ds
+              = A.Clause (A.LHS i x aps (pats ++ ps)) (insertPatterns pats rhs) ds
+insertPatterns pats (A.RewriteRHS qs eqs rhs wh) = A.RewriteRHS qs eqs (insertPatterns pats rhs) wh
+insertPatterns pats rhs = rhs
+
+
 data WithFunctionProblem
       = NoWithFunction
       | WithFunction QName          -- parent function name
@@ -129,11 +142,12 @@ data WithFunctionProblem
 
 -- | Type check a function clause.
 checkClause :: Type -> A.Clause -> TCM Clause
-checkClause t c@(A.Clause (A.LHS i x aps []) rhs wh) =
+checkClause t c@(A.Clause (A.LHS i x aps []) rhs0 wh) =
     traceCall (CheckClause t c) $
     checkLeftHandSide c aps t $ \gamma delta sub xs ps t' perm -> do
       let mkBody v = foldr (\x t -> Bind $ Abs x t) (Body $ substs sub v) xs
-      (body, with) <- checkWhere (size delta) wh $
+      (body, with) <- checkWhere (size delta) wh $ let
+          handleRHS rhs =
               case rhs of
                 A.RHS e
                   | any (containsAbsurdPattern . namedThing . unArg) aps ->
@@ -145,8 +159,56 @@ checkClause t c@(A.Clause (A.LHS i x aps []) rhs wh) =
                   | any (containsAbsurdPattern . namedThing . unArg) aps
                               -> return (NoBody, NoWithFunction)
                   | otherwise -> typeError $ NoRHSRequiresAbsurdPattern aps
+                A.RewriteRHS [] (_:_) _ _ -> __IMPOSSIBLE__
+                A.RewriteRHS (_:_) [] _ _ -> __IMPOSSIBLE__
+                A.RewriteRHS [] [] rhs [] -> handleRHS rhs
+                A.RewriteRHS [] [] _ (_:_) -> __IMPOSSIBLE__
+                A.RewriteRHS (qname:names) (eq:eqs) rhs wh -> do
+                     (proof,t) <- inferExpr eq
+                     t' <- reduce =<< instantiateFull t
+                     equality <- primEquality >>= \eq -> return $ case eq of
+                        Lam Hidden (Abs _ (Def equality _)) -> equality
+                        _                                   -> __IMPOSSIBLE__
+                     reflCon <- primRefl >>= \refl -> return $ case refl of
+                         Lam Hidden (Abs _typ (Lam Hidden (Abs _val (Con reflCon [])))) -> reflCon
+                         _ -> __IMPOSSIBLE__
+                     (rewriteType,rewriteFrom,rewriteTo) <- case t' of
+                         El _Set0 (Def equality' [Arg Hidden rewriteType,
+                                                  Arg NotHidden rewriteFrom, Arg NotHidden rewriteTo])
+                            | equality' == equality ->
+                              return (rewriteType, rewriteFrom, rewriteTo)
+                         -- Universe polymorphic equality
+                         El _Set0 (Def equality' [_level,Arg Hidden rewriteType,
+                                                  Arg NotHidden rewriteFrom, Arg NotHidden rewriteTo])
+                            | equality' == equality ->
+                              return (rewriteType, rewriteFrom, rewriteTo)
+                         _ -> do
+                          err <- text "Cannot rewrite by equation of type" <+> prettyTCM t'
+                          typeError $ GenericError $ show err
+                         
+                     let info = PatRange noRange
+                         metaInfo = Info.MetaInfo noRange emptyScopeInfo Nothing
+                         underscore = A.Underscore metaInfo
+                         
+                     [rewriteFromExpr,rewriteToExpr,rewriteTypeExpr, proofExpr] <- setShowImplicitArguments True $ reify
+                      [rewriteFrom,   rewriteTo,    rewriteType    , proof]
+                     let (inner, outer) -- the where clauses should go on the inner-most with
+                           | null eqs  = ([], wh)
+                           | otherwise = (wh, [])
+                         newRhs = A.WithRHS qname [rewriteFromExpr, proofExpr] 
+                                  [A.Clause (A.LHS i x aps pats)
+                                    (A.RewriteRHS names eqs (insertPatterns pats rhs) inner)
+                                    outer]
+                         pats = [A.DotP info underscore, -- rewriteToExpr,
+                                 A.ConP info (AmbQ [reflCon]) []]
+                     reportSDoc "tc.rewrite.top" 25 $ vcat 
+                                         [ text "from = " <+> prettyTCM rewriteFromExpr,
+                                           text "to = " <+> prettyTCM rewriteToExpr,
+                                           text "typ = " <+> prettyTCM rewriteType,
+                                           text "proof = " <+> prettyTCM proofExpr,
+                                           text "equ = " <+> prettyTCM t' ]
+                     handleRHS newRhs
                 A.WithRHS aux es cs -> do
-
                   -- Infer the types of the with expressions
                   vas <- mapM inferExpr es
                   (vs, as) <- instantiateFull $ unzip vas
@@ -195,6 +257,7 @@ checkClause t c@(A.Clause (A.LHS i x aps []) rhs wh) =
                     ]
 
                   return (mkBody v, WithFunction x aux gamma delta1 delta2 vs as t' ps finalPerm cs)
+          in handleRHS rhs0
       escapeContext (size delta) $ checkWithFunction with
 
       reportSDoc "tc.lhs.top" 10 $ vcat
@@ -293,8 +356,8 @@ checkWhere n [A.Section _ m tel ds]  ret = do
       dx   <- prettyTCM m
       dtel <- mapM prettyA tel
       dtel' <- prettyTCM =<< lookupSection m
-      liftIO $ UTF8.putStrLn $ "checking where section " ++ show dx ++ " " ++ show dtel
-      liftIO $ UTF8.putStrLn $ "        actual tele: " ++ show dtel'
+      liftIO $ LocIO.putStrLn $ "checking where section " ++ show dx ++ " " ++ show dtel
+      liftIO $ LocIO.putStrLn $ "        actual tele: " ++ show dtel'
     x <- withCurrentModule m $ checkDecls ds >> ret
     return x
 checkWhere _ _ _ = __IMPOSSIBLE__
@@ -320,7 +383,7 @@ checkLHS ps t ret = do
     verbose 15 $ do
       dt  <- prettyTCM t
       dps <- mapM prettyA ps
-      liftIO $ UTF8.putStrLn $ "checking clause " ++ show dps ++ " : " ++ show dt
+      liftIO $ LocIO.putStrLn $ "checking clause " ++ show dps ++ " : " ++ show dt
 
     -- Save the state for later. (should this be done with the undo monad, or
     -- would that interfere with normal undo?)
@@ -341,18 +404,18 @@ checkLHS ps t ret = do
         d0 <- A.showA ps0
         d1 <- A.showA ps1
         liftIO $ do
-        UTF8.putStrLn $ "first check"
-        UTF8.putStrLn $ "  xs    = " ++ show xs
-        UTF8.putStrLn $ "  metas = " ++ show metas
-        UTF8.putStrLn $ "  ps0   = " ++ d0
-        UTF8.putStrLn $ "  ps1   = " ++ d1
+        LocIO.putStrLn $ "first check"
+        LocIO.putStrLn $ "  xs    = " ++ show xs
+        LocIO.putStrLn $ "  metas = " ++ show metas
+        LocIO.putStrLn $ "  ps0   = " ++ d0
+        LocIO.putStrLn $ "  ps1   = " ++ d1
 
     verbose 10 $ do
         is <- mapM (instantiateFull . flip MetaV []) metas
         ds <- mapM prettyTCM is
         dts <- mapM prettyTCM =<< mapM instantiateFull ts
-        liftIO $ UTF8.putStrLn $ "  is    = " ++ concat (intersperse ", " $ map show ds)
-        liftIO $ UTF8.putStrLn $ "  ts    = " ++ concat (intersperse ", " $ map show dts)
+        liftIO $ LocIO.putStrLn $ "  is    = " ++ concat (intersperse ", " $ map show ds)
+        liftIO $ LocIO.putStrLn $ "  ts    = " ++ concat (intersperse ", " $ map show dts)
 
     -- Now we forget that we ever type checked anything and type check the new
     -- pattern.
@@ -363,14 +426,14 @@ checkLHS ps t ret = do
     mapM_ isEmptyType emptyTypes
 
     verbose 10 $ liftIO $ do
-        UTF8.putStrLn $ "second check"
-        UTF8.putStrLn $ "  xs    = " ++ show xs
-        UTF8.putStrLn $ "  metas = " ++ show metas
+        LocIO.putStrLn $ "second check"
+        LocIO.putStrLn $ "  xs    = " ++ show xs
+        LocIO.putStrLn $ "  metas = " ++ show metas
 
     verbose 10 $ do
         is <- mapM (instantiateFull . flip MetaV []) metas
         ds <- mapM prettyTCM is
-        liftIO $ UTF8.putStrLn $ "  is    = " ++ concat (intersperse ", " $ map show ds)
+        liftIO $ LocIO.putStrLn $ "  is    = " ++ concat (intersperse ", " $ map show ds)
 
     -- Finally we type check the dot patterns and check that they match their
     -- instantiations.
@@ -412,8 +475,8 @@ checkLHS ps t ret = do
         verbose 20 $ do
             d <- prettyTCM ctx
             dt <- prettyTCM (substs sub a)
-            liftIO $ UTF8.putStrLn $ "context = " ++ show d
-            liftIO $ UTF8.putStrLn $ "type    = " ++ show dt
+            liftIO $ LocIO.putStrLn $ "context = " ++ show d
+            liftIO $ LocIO.putStrLn $ "type    = " ++ show dt
 
         reportLn 20 $ "finished type checking left hand side"
         ret rsub xs ps (substs sub a)
@@ -434,7 +497,7 @@ checkLHS ps t ret = do
             v <- lift $ instantiate (MetaV x [])
             lift $ verbose 6 $ do
                 d <- prettyTCM v
-                liftIO $ UTF8.putStrLn $ "new pattern for " ++ show x ++ " = " ++ show d
+                liftIO $ LocIO.putStrLn $ "new pattern for " ++ show x ++ " = " ++ show d
             case v of
                 -- Unsolved metas become variables
                 MetaV y _ | x == y  -> return $ A.WildP i
@@ -442,7 +505,7 @@ checkLHS ps t ret = do
                 _                   -> do
                     lift $ verbose 6 $ do
                         d <- prettyTCM =<< instantiateFull v
-                        liftIO $ UTF8.putStrLn $ show x ++ " := " ++ show d
+                        liftIO $ LocIO.putStrLn $ show x ++ " := " ++ show d
                     scope <- lift getScope
                     return $ A.DotP i (A.Underscore $ info scope)
                     where info s = Info.MetaInfo (getRange i) s Nothing
@@ -506,7 +569,7 @@ checkLHS ps t ret = do
                     xs   <- free t
                     verbose 20 $ do
                         d <- prettyTCM t
-                        liftIO $ UTF8.putStrLn $ "freeIn " ++ show x ++ " : " ++ show d ++ " are " ++ show xs
+                        liftIO $ LocIO.putStrLn $ "freeIn " ++ show x ++ " : " ++ show d ++ " are " ++ show xs
                     case intersect (map (fst . unArg) tel') xs of
                         [] -> return $ Arg h (x,t) : tel'
                         zs -> return $ ins zs (Arg h (x,t)) tel'
@@ -543,7 +606,7 @@ checkLHS ps t ret = do
                 par Hidden    s = "{" ++ s ++ "}"
                 par NotHidden s = "(" ++ s ++ ")"
             ds <- mapM pr ctx
-            liftIO $ UTF8.putStr $ unlines $ reverse ds
+            liftIO $ LocIO.putStr $ unlines $ reverse ds
 -}
 
 actualConstructor :: MonadTCM tcm => QName -> tcm QName
