@@ -8,8 +8,9 @@ import Control.Monad.Reader
 import Control.Monad.Error
 import Data.Maybe
 import Data.List hiding (sort)
-import qualified System.IO.UTF8 as UTF8
+import qualified Agda.Utils.IO.Locale as LocIO
 import qualified Data.Map as Map
+import qualified Data.Set as Set
 
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Abstract.Views as A
@@ -32,6 +33,7 @@ import Agda.TypeChecking.Records
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.Constraints
+import Agda.TypeChecking.Free
 
 import Agda.Utils.Fresh
 import Agda.Utils.Tuple
@@ -133,12 +135,13 @@ checkLiteral lit t = do
 
 litType :: Literal -> TCM Type
 litType l = case l of
+    LitLevel _ _  -> el <$> primLevel
     LitInt _ _	  -> el <$> primNat
     LitFloat _ _  -> el <$> primFloat
     LitChar _ _   -> el <$> primChar
     LitString _ _ -> el <$> primString
   where
-    el t = El (Type 0) t
+    el t = El (mkType 0) t
 
 ---------------------------------------------------------------------------
 -- * Terms
@@ -177,6 +180,10 @@ checkExpr e t =
         text "    --> " <+> prettyTCM t
     let scopedExpr (A.ScopedExpr scope e) = setScope scope >> scopedExpr e
 	scopedExpr e			  = return e
+
+        unScope (A.ScopedExpr scope e) = unScope e
+        unScope e                      = e
+
     e <- scopedExpr e
     case e of
 
@@ -220,13 +227,14 @@ checkExpr e t =
                         text "target type: " <+> prettyTCM t1
                       case t1 of
                         NotBlocked (Def d _) -> do
+                          let dataOrRec = case [ c | (d', c) <- dcs, d == d' ] of
+                                c:_ -> return (Just c)
+                                []  -> typeError $ DoesNotConstructAnElementOf
+                                                     (head cs) (Def d [])
                           defn <- theDef <$> getConstInfo d
                           case defn of
-                            Datatype{} ->
-                              case [ c | (d', c) <- dcs, d == d' ] of
-                                c:_   -> return (Just c)
-                                []    -> typeError $ DoesNotConstructAnElementOf
-                                          (head cs) (Def d [])
+                            Datatype{} -> dataOrRec
+                            Record{}   -> dataOrRec
                             _ -> typeError $ DoesNotConstructAnElementOf (head cs) (ignoreBlocking t1)
                         NotBlocked (MetaV _ _)  -> return Nothing
                         Blocked{} -> return Nothing
@@ -241,6 +249,19 @@ checkExpr e t =
 
 	A.WithApp _ e es -> typeError $ NotImplemented "type checking of with application"
 
+        A.App i s (Arg NotHidden l)
+          | A.Set _ 0 <- unScope s ->
+          ifM (not <$> hasUniversePolymorphism)
+              (typeError $ GenericError "Use --universe-polymorphism to enable level arguments to Set")
+          $ do
+            lvl <- primLevel
+            suc <- do s <- primLevelSuc
+                      return $ \x -> s `apply` [Arg NotHidden x]
+            n   <- checkExpr (namedThing l) (El (mkType 0) lvl)
+            reportSDoc "tc.univ.poly" 10 $
+              text "checking Set " <+> prettyTCM n <+> text "against" <+> prettyTCM t
+            blockTerm t (Sort $ Type n) $ leqType (sort $ Type $ suc n) t
+
 	A.App i e arg -> do
 	    (v0, t0)	 <- inferExpr e
 	    checkArguments' ExpandLast (getRange e) [arg] t0 t e $ \vs t1 cs ->
@@ -329,8 +350,17 @@ checkExpr e t =
 	A.Lit lit    -> checkLiteral lit t
 	A.Let i ds e -> checkLetBindings ds $ checkExpr e t
 	A.Pi _ tel e -> do
-	    t' <- checkTelescope tel $ \tel -> telePi_ tel <$> isType_ e
-	    blockTerm t (unEl t') $ leqType (sort $ getSort t') t
+	    t' <- checkTelescope tel $ \tel -> do
+                    t   <- instantiateFull =<< isType_ e
+                    tel <- instantiateFull tel
+                    return $ telePi_ tel t
+            s  <- return $ getSort t'
+            when (s == Inf) $ reportSDoc "tc.term.sort" 20 $
+              vcat [ text ("reduced to omega:")
+                   , nest 2 $ text "t   =" <+> prettyTCM t'
+                   , nest 2 $ text "cxt =" <+> (prettyTCM =<< getContextTelescope)
+                   ]
+	    blockTerm t (unEl t') $ leqType (sort s) t
 	A.Fun _ (Arg h a) b -> do
 	    a' <- isType_ a
 	    b' <- isType_ b
@@ -338,24 +368,25 @@ checkExpr e t =
 	    blockTerm t (Fun (Arg h a') b') $ leqType (sort s) t
 	A.Set _ n    -> do
           n <- ifM typeInType (return 0) (return n)
-	  blockTerm t (Sort (Type n)) $ leqType (sort $ Type $ n + 1) t
+	  blockTerm t (Sort (mkType n)) $ leqType (sort $ mkType $ n + 1) t
 	A.Prop _     -> do
-          s <- ifM typeInType (return $ Type 0) (return Prop)
-	  blockTerm t (Sort Prop) $ leqType (sort $ Type 1) t
+          s <- ifM typeInType (return $ mkType 0) (return Prop)
+	  blockTerm t (Sort Prop) $ leqType (sort $ mkType 1) t
 
 	A.Rec _ fs  -> do
 	  t <- normalise t
 	  case unEl t of
 	    Def r vs  -> do
-	      xs    <- getRecordFieldNames r
-	      ftel  <- getRecordFieldTypes r
-              scope <- getScope
+	      (hs, xs) <- unzip <$> getRecordFieldNames r
+	      ftel     <- getRecordFieldTypes r
+              con      <- getRecordConstructor r
+              scope    <- getScope
               let meta = A.Underscore $ A.MetaInfo (getRange e) scope Nothing
 	      es   <- orderFields r meta xs fs
 	      let tel = ftel `apply` vs
 	      (args, cs) <- checkArguments_ ExpandLast (getRange e)
-			      (map (Arg NotHidden . unnamed) es) tel
-	      blockTerm t (Con (killRange r) args) $ return cs
+			      (zipWith (\h e -> Arg h (unnamed e)) hs es) tel
+	      blockTerm t (Con con args) $ return cs
             MetaV _ _ -> do
               reportSDoc "tc.term.expr.rec" 10 $ sep
                 [ text "Postponing type checking of"
@@ -393,7 +424,7 @@ inferHead (HeadCon [c]) = do
   Constructor{conPars = n} <- theDef <$> (instantiateDef =<< getConstInfo c)
 
   verboseS "tc.term.con" 7 $ do
-    liftIO $ UTF8.putStrLn $ unwords [show c, "has", show n, "parameters."]
+    liftIO $ LocIO.putStrLn $ unwords [show c, "has", show n, "parameters."]
 
   -- So when applying the constructor throw away the parameters.
   return (apply u . genericDrop n, a)
@@ -408,7 +439,7 @@ inferDef mkTerm x =
       ds <- mapM prettyTCM vs
       dx <- prettyTCM x
       dt <- prettyTCM $ defType d
-      liftIO $ UTF8.putStrLn $ "inferred def " ++ unwords (show dx : map show ds) ++ " : " ++ show dt
+      liftIO $ LocIO.putStrLn $ "inferred def " ++ unwords (show dx : map show ds) ++ " : " ++ show dt
     return (mkTerm x vs, defType d)
 
 -- | @checkHeadApplication e t hd args@ checks that @e@ has type @t@,
