@@ -1,5 +1,6 @@
 {-# LANGUAGE OverlappingInstances,
              TypeSynonymInstances,
+             IncoherentInstances,
              ExistentialQuantification,
              ScopedTypeVariables,
              CPP
@@ -13,9 +14,6 @@
 -- version number should be bumped _in the same patch_.
 -- -!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-
 
--- TODO: It should be easy to produce a decent QuickCheck test suite
--- for this file.
-
 module Agda.TypeChecking.Serialise
   ( encode
   , encodeFile
@@ -25,11 +23,14 @@ module Agda.TypeChecking.Serialise
   )
   where
 
+import qualified Control.Exception as E
 import Control.Monad
-import Control.Monad.State.Strict
 import Control.Monad.Reader
+import Control.Monad.State.Strict
+import Control.Monad.Error
 import Data.Array.IArray
 import Data.Bits (shiftR)
+import Data.Word
 import Data.ByteString.Lazy as L
 import Data.Char (ord, chr)
 import Data.HashTable (HashTable)
@@ -38,6 +39,8 @@ import Data.Int (Int32, Int64)
 import Data.IORef
 import Data.Map (Map)
 import qualified Data.Map as M
+import Data.Set (Set)
+import qualified Data.Set as S
 import qualified Data.Binary as B
 import qualified Data.Binary.Get as B
 import qualified Data.List as List
@@ -58,16 +61,23 @@ import Agda.Syntax.Fixity
 import Agda.Syntax.Literal
 import qualified Agda.Interaction.Highlighting.Range   as HR
 import qualified Agda.Interaction.Highlighting.Precise as HP
+import Agda.Interaction.FindFile
 
 import Agda.TypeChecking.Monad
+import Agda.Utils.FileName
+import Agda.Utils.Monad
 import Agda.Utils.Tuple
 import Agda.Utils.Permutation
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
-currentInterfaceVersion :: Int
-currentInterfaceVersion = 20090421 * 10 + 2
+-- Note that the Binary instance for Int writes 64 bits, but throws
+-- away the 32 high bits when reading (at the time of writing, on
+-- 32-bit machines). Word64 does not have these problems.
+
+currentInterfaceVersion :: Word64
+currentInterfaceVersion = 20091123 * 10 + 0
 
 type Node = [Int] -- constructor tag (maybe omitted) and arg indices
 
@@ -79,57 +89,135 @@ data Dict = Dict{ nodeD     :: !(HashTable Node    Int)
                 , stringC   :: !(IORef Int)
                 , integerC  :: !(IORef Int)
                 , doubleC   :: !(IORef Int)
+                , fileMod   :: !SourceToModule
                 }
 
 data U    = forall a . Data a => U !a
 type Memo = HashTable (Int, Int) U    -- (node index, type rep key)
 
-data Env  = Env { nodeE     :: !(Array Int Node)
-                , stringE   :: !(Array Int String)
-                , integerE  :: !(Array Int Integer)
-                , doubleE   :: !(Array Int Double)
-                , nodeMemo  :: !Memo
-                }
+data St = St
+  { nodeE     :: !(Array Int Node)
+  , stringE   :: !(Array Int String)
+  , integerE  :: !(Array Int Integer)
+  , doubleE   :: !(Array Int Double)
+  , nodeMemo  :: !Memo
+  , modFile   :: !ModuleToSource
+    -- ^ Maps module names to file names. This is the only component
+    -- of the state which is updated by the decoder.
+  , includes  :: [AbsolutePath]
+    -- ^ The include directories.
+  }
+
+-- | Monad used by the encoder.
 
 type S a = ReaderT Dict IO a
-type R a = ReaderT Env  IO a
+
+-- | Monad used by the decoder.
+--
+-- 'TCM' is not used because the associated overheads would make
+-- decoding slower.
+
+type R a = ErrorT TypeError (StateT St IO) a
+
+-- | Throws an error which is suitable when the data stream is
+-- malformed.
+
+malformed :: R a
+malformed = throwError $ GenericError "Malformed input."
 
 class Data a => EmbPrj a where
   icode :: a -> S Int
   value :: Int -> R a
 
+-- | Encodes something. To ensure relocatability file paths in
+-- positions are replaced with module names.
 
-encode :: EmbPrj a => a -> IO ByteString
+encode :: EmbPrj a => a -> TCM ByteString
 encode a = do
-    newD@(Dict nD sD iD dD _ _ _ _) <- emptyDict
-    root <- runReaderT (icode a) newD
-    nL <- l nD; sL <- l sD; iL <- l iD; dL <- l dD
-    return $ B.encode currentInterfaceVersion `L.append`
-             G.compress (B.encode (root, nL, sL, iL, dL))
+    fileMod <- sourceToModule
+    liftIO $ do
+      newD@(Dict nD sD iD dD _ _ _ _ _) <- emptyDict fileMod
+      root <- runReaderT (icode a) newD
+      nL <- l nD; sL <- l sD; iL <- l iD; dL <- l dD
+      return $ B.encode currentInterfaceVersion `L.append`
+               G.compress (B.encode (root, nL, sL, iL, dL))
   where l = fmap (List.map fst . List.sortBy (compare `on` snd)) . H.toList
 
-decode :: EmbPrj a => ByteString -> IO a
-decode s | ver /= currentInterfaceVersion = error "Wrong interface version"
-         | otherwise = runReaderT (value r) . env =<< H.new (==) hashInt2
-  where (ver                , s1, _) = B.runGetState B.get s                 0
-        ((r, nL, sL, iL, dL), s2, _) = B.runGetState B.get (G.decompress s1) 0
-        ar l = listArray (0, List.length l - 1) l
-        env  = Env (ar nL) (ar sL) (ar iL) (ar dL)
+-- | Decodes something. The result depends on the include path.
+--
+-- Returns 'Nothing' if the input does not start with the right magic
+-- number or some other decoding error is encountered.
 
-encodeFile :: EmbPrj a => FilePath -> a -> IO ()
-encodeFile f x = L.writeFile f =<< encode x
+decode :: EmbPrj a => ByteString -> TCM (Maybe a)
+decode s = do
+  mf   <- stModuleToSource <$> get
+  incs <- getIncludeDirs
 
-decodeFile :: EmbPrj a => FilePath -> IO a
-decodeFile f = decode =<< L.readFile f
+  -- Note that B.runGetState and G.decompress can raise errors if the
+  -- input is malformed. The decoder is (intended to be) strict enough
+  -- to ensure that all such errors can be caught by the handler here.
+
+  (mf, x) <- liftIO $ E.handle (\(E.ErrorCall {}) -> noResult) $ do
+
+    (ver, s, _) <- return $ B.runGetState B.get s 0
+    if ver /= currentInterfaceVersion
+     then noResult
+     else do
+
+      ((r, nL, sL, iL, dL), s, _) <-
+        return $ B.runGetState B.get (G.decompress s) 0
+      if s /= L.empty
+         -- G.decompress seems to throw away garbage at the end, so
+         -- the then branch is possibly dead code.
+       then noResult
+       else do
+
+        st <- St (ar nL) (ar sL) (ar iL) (ar dL)
+                <$> liftIO (H.new (==) hashInt2)
+                <*> return mf <*> return incs
+        (r, st) <- runStateT (runErrorT (value r)) st
+        return (Just (modFile st), case r of
+          Left  _ -> Nothing
+          Right x -> Just x)
+
+  case mf of
+    Nothing -> return ()
+    Just mf -> modify $ \s -> s { stModuleToSource = mf }
+
+  return x
+
+  where
+  ar l = listArray (0, List.length l - 1) l
+
+  noResult = return (Nothing, Nothing)
+
+-- | Encodes something. To ensure relocatability file paths in
+-- positions are replaced with module names.
+
+encodeFile :: EmbPrj a
+           => FilePath
+              -- ^ The encoded data is written to this file.
+           -> a
+              -- ^ Something.
+           -> TCM ()
+encodeFile f x = liftIO . L.writeFile f =<< encode x
+
+-- | Decodes something. The result depends on the include path.
+--
+-- Returns 'Nothing' if the file does not start with the right magic
+-- number or some other decoding error is encountered.
+
+decodeFile :: EmbPrj a => FilePath -> TCM (Maybe a)
+decodeFile f = decode =<< liftIO (L.readFile f)
 
 
 instance EmbPrj String where
   icode   = icodeX stringD stringC
-  value i = (! i) `fmap` asks stringE
+  value i = (! i) `fmap` gets stringE
 
 instance EmbPrj Integer where
   icode   = icodeX integerD integerC
-  value i = (! i) `fmap` asks integerE
+  value i = (! i) `fmap` gets integerE
 
 instance EmbPrj Int where
   icode i = return i
@@ -141,36 +229,59 @@ instance EmbPrj Char where
 
 instance EmbPrj Double where
   icode   = icodeX doubleD doubleC
-  value i = (! i) `fmap` asks doubleE
+  value i = (! i) `fmap` gets doubleE
 
 instance (EmbPrj a, EmbPrj b) => EmbPrj (a, b) where
   icode (a, b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 (,) a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance (EmbPrj a, EmbPrj b, EmbPrj c) => EmbPrj (a, b, c) where
   icode (a, b, c) = icode3' a b c
   value = vcase valu where valu [a, b, c] = valu3 (,,) a b c
-                           valu _         = __IMPOSSIBLE__
+                           valu _         = malformed
 
 instance EmbPrj a => EmbPrj (Maybe a) where
   icode Nothing  = icode0'
   icode (Just x) = icode1' x
   value = vcase valu where valu []  = valu0 Nothing
                            valu [x] = valu1 Just x
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
 
 instance EmbPrj Bool where
   icode True  = icode0 0
   icode False = icode0 1
   value = vcase valu where valu [0] = valu0 True
                            valu [1] = valu0 False
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
+
+instance EmbPrj AbsolutePath where
+  icode file = do
+    mm <- M.lookup file . fileMod <$> ask
+    case mm of
+      Just m  -> icode m
+      Nothing -> __IMPOSSIBLE__
+  value m = do
+    m       <- value m
+    mf      <- modFile  <$> get
+    incs    <- includes <$> get
+    (r, mf) <- liftIO $ findFile'' incs m mf
+    modify $ \s -> s { modFile = mf }
+    case r of
+      Left err -> throwError $ findErrorToTypeError m err
+      Right f  -> return f
 
 instance EmbPrj Position where
   icode (P.Pn file pos line col) = icode4' file pos line col
-  value = vcase valu where valu [f, p, l, c] = valu4 P.Pn f p l c
-                           valu _            = __IMPOSSIBLE__
+  value = vcase valu
+    where
+    valu [f, p, l, c] = valu4 P.Pn f p l c
+    valu _            = malformed
+
+instance EmbPrj TopLevelModuleName where
+  icode (TopLevelModuleName a) = icode1' a
+  value = vcase valu where valu [a] = valu1 TopLevelModuleName a
+                           valu _   = malformed
 
 instance EmbPrj a => EmbPrj [a] where
   icode xs = icodeN =<< mapM icode xs
@@ -179,69 +290,73 @@ instance EmbPrj a => EmbPrj [a] where
 --   icode (x : xs) = icode2' x xs
 --   value = vcase valu where valu []      = valu0 []
 --                            valu [x, xs] = valu2 (:) x xs
---                            valu _       = __IMPOSSIBLE__
+--                            valu _       = malformed
 
 instance (Ord a, EmbPrj a, EmbPrj b) => EmbPrj (Map a b) where
   icode m = icode (M.toList m)
   value m = M.fromList `fmap` value m
 
+instance (Ord a, EmbPrj a) => EmbPrj (Set a) where
+  icode s = icode (S.toList s)
+  value s = S.fromList `fmap` value s
+
 instance EmbPrj P.Interval where
   icode (P.Interval p q) = icode2' p q
   value = vcase valu where valu [p, q] = valu2 P.Interval p q
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj Range where
   icode (P.Range is) = icode1' is
   value = vcase valu where valu [is] = valu1 P.Range is
-                           valu _    = __IMPOSSIBLE__
+                           valu _    = malformed
 
 instance EmbPrj HR.Range where
   icode (HR.Range a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 HR.Range a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj C.Name where
   icode (C.NoName a b) = icode2 0 a b
   icode (C.Name r xs)  = icode2 1 r xs
   value = vcase valu where valu [0, a, b]  = valu2 C.NoName a b
                            valu [1, r, xs] = valu2 C.Name   r xs
-                           valu _          = __IMPOSSIBLE__
+                           valu _          = malformed
 
 instance EmbPrj NamePart where
   icode Hole   = icode0'
   icode (Id a) = icode1' a
   value = vcase valu where valu []  = valu0 Hole
                            valu [a] = valu1 Id a
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
 
 instance EmbPrj C.QName where
   icode (Qual    a b) = icode2' a b
   icode (C.QName a  ) = icode1' a
   value = vcase valu where valu [a, b] = valu2 Qual    a b
                            valu [a]    = valu1 C.QName a
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj Scope where
   icode (Scope a b c d e f) = icode6' a b c d e f
   value = vcase valu where valu [a, b, c, d, e, f] = valu6 Scope a b c d e f
-                           valu _                  = __IMPOSSIBLE__
+                           valu _                  = malformed
 
 instance EmbPrj Access where
   icode PrivateAccess = icode0 0
   icode PublicAccess  = icode0 1
   value = vcase valu where valu [0] = valu0 PrivateAccess
                            valu [1] = valu0 PublicAccess
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
 
 instance EmbPrj NameSpace where
   icode (NameSpace a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 NameSpace a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj AbstractName where
   icode (AbsName a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 AbsName a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj AbstractModule where
   icode (AbsModule a) = icode a
@@ -252,7 +367,7 @@ instance EmbPrj KindOfName where
   icode ConName = icode0 1
   value = vcase valu where valu [0] = valu0 DefName
                            valu [1] = valu0 ConName
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
 
 instance EmbPrj Agda.Syntax.Fixity.Fixity where
   icode (LeftAssoc  a b) = icode2 0 a b
@@ -261,12 +376,12 @@ instance EmbPrj Agda.Syntax.Fixity.Fixity where
   value = vcase valu where valu [0, a, b] = valu2 LeftAssoc  a b
                            valu [1, a, b] = valu2 RightAssoc a b
                            valu [2, a, b] = valu2 NonAssoc   a b
-                           valu _         = __IMPOSSIBLE__
+                           valu _         = malformed
 
 instance EmbPrj A.QName where
   icode (A.QName a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 A.QName a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj A.ModuleName where
   icode (A.MName a) = icode a
@@ -275,58 +390,58 @@ instance EmbPrj A.ModuleName where
 instance EmbPrj A.Name where
   icode (A.Name a b c d) = icode4' a b c d
   value = vcase valu where valu [a, b, c, d] = valu4 A.Name a b c d
-                           valu _            = __IMPOSSIBLE__
+                           valu _            = malformed
 
 instance EmbPrj NameId where
   icode (NameId a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 NameId a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj Signature where
   icode (Sig a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 Sig a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj Section where
   icode (Section a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 Section a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj Telescope where
   icode EmptyTel        = icode0'
   icode (ExtendTel a b) = icode2' a b
   value = vcase valu where valu []     = valu0 EmptyTel
                            valu [a, b] = valu2 ExtendTel a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj Permutation where
   icode (Perm a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 Perm a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance (EmbPrj a) => EmbPrj (Agda.Syntax.Common.Arg a) where
   icode (Arg a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 Arg a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj Agda.Syntax.Common.Induction where
   icode Inductive   = icode0 0
   icode CoInductive = icode0 1
   value = vcase valu where valu [0] = valu0 Inductive
                            valu [1] = valu0 CoInductive
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
 
 instance EmbPrj Agda.Syntax.Common.Hiding where
   icode Hidden    = icode0 0
   icode NotHidden = icode0 1
   value = vcase valu where valu [0] = valu0 Hidden
                            valu [1] = valu0 NotHidden
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
 
 instance EmbPrj I.Type where
   icode (El a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 El a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj I.MetaId where
   icode (MetaId a) = icode a
@@ -335,7 +450,7 @@ instance EmbPrj I.MetaId where
 instance (EmbPrj a) => EmbPrj (I.Abs a) where
   icode (Abs a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 Abs a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj I.Term where
   icode (Var      a b) = icode2 0 a b
@@ -356,41 +471,47 @@ instance EmbPrj I.Term where
                            valu [6, a, b] = valu2 Fun   a b
                            valu [7, a]    = valu1 Sort  a
                            valu [8, a, b] = valu2 MetaV a b
-                           valu _         = __IMPOSSIBLE__
+                           valu _         = malformed
 
 instance EmbPrj I.Sort where
   icode (Type  a  ) = icode1 0 a
   icode Prop        = icode0 1
   icode (Lub   a b) = icode2 2 a b
   icode (Suc   a  ) = icode1 3 a
-  icode (MetaS a )  = icode1 4 a
+  icode (MetaS a b) = icode2 4 a b
+  icode Inf         = icode0 5
+  icode (DLub a b)  = icode2 6 a b
   value = vcase valu where valu [0, a]    = valu1 Type  a
                            valu [1]       = valu0 Prop
                            valu [2, a, b] = valu2 Lub   a b
                            valu [3, a]    = valu1 Suc   a
-                           valu [4, a]    = valu1 MetaS a
-                           valu _         = __IMPOSSIBLE__
+                           valu [4, a, b] = valu2 MetaS a b
+                           valu [5]       = valu0 Inf
+                           valu [6, a, b] = valu2 DLub a b
+                           valu _         = malformed
 
 instance EmbPrj Agda.Syntax.Literal.Literal where
   icode (LitInt    a b) = icode2 0 a b
   icode (LitFloat  a b) = icode2 1 a b
   icode (LitString a b) = icode2 2 a b
   icode (LitChar   a b) = icode2 3 a b
+  icode (LitLevel  a b) = icode2 4 a b
   value = vcase valu where valu [0, a, b] = valu2 LitInt    a b
                            valu [1, a, b] = valu2 LitFloat  a b
                            valu [2, a, b] = valu2 LitString a b
                            valu [3, a, b] = valu2 LitChar   a b
-                           valu _         = __IMPOSSIBLE__
+                           valu [4, a, b] = valu2 LitLevel  a b
+                           valu _         = malformed
 
 instance EmbPrj DisplayForm where
   icode (Display a b c) = icode3' a b c
   value = vcase valu where valu [a, b, c] = valu3 Display a b c
-                           valu _         = __IMPOSSIBLE__
+                           valu _         = malformed
 
 instance EmbPrj a => EmbPrj (Open a) where
   icode (OpenThing a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 OpenThing a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj CtxId where
   icode (CtxId a) = icode a
@@ -401,7 +522,7 @@ instance EmbPrj DisplayTerm where
   icode (DWithApp a b) = icode2' a b
   value = vcase valu where valu [a]    = valu1 DTerm a
                            valu [a, b] = valu2 DWithApp a b
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj MutualId where
   icode (MutId a) = icode a
@@ -410,7 +531,7 @@ instance EmbPrj MutualId where
 instance EmbPrj Definition where
   icode (Defn a b c d e) = icode5' a b c d e
   value = vcase valu where valu [a, b, c, d, e] = valu5 Defn a b c d e
-                           valu _               = __IMPOSSIBLE__
+                           valu _               = malformed
 
 instance EmbPrj HaskellRepresentation where
   icode (HsType a)   = icode1 0 a
@@ -419,7 +540,7 @@ instance EmbPrj HaskellRepresentation where
   value = vcase valu where
     valu [0, a]    = valu1 HsType a
     valu [1, a, b] = valu2 HsDefn a b
-    valu _         = __IMPOSSIBLE__
+    valu _         = malformed
 
 instance EmbPrj Polarity where
   icode Covariant     = icode0 0
@@ -430,7 +551,7 @@ instance EmbPrj Polarity where
     valu [0] = valu0 Covariant
     valu [1] = valu0 Contravariant
     valu [2] = valu0 Invariant
-    valu _   = __IMPOSSIBLE__
+    valu _   = malformed
 
 instance EmbPrj Occurrence where
   icode Positive = icode0 0
@@ -441,30 +562,30 @@ instance EmbPrj Occurrence where
     valu [0] = valu0 Positive
     valu [1] = valu0 Negative
     valu [2] = valu0 Unused
-    valu _   = __IMPOSSIBLE__
+    valu _   = malformed
 
 instance EmbPrj Defn where
   icode (Axiom       a)                   = icode1 0 a
   icode (Function    a b c d e f)         = icode6 1 a b c d e f
   icode (Datatype    a b c d e f g h i j) = icode10 2 a b c d e f g h i j
-  icode (Record      a b c d e f g h)     = icode8 3 a b c d e f g h
+  icode (Record      a b c d e f g h i)   = icode9 3 a b c d e f g h i
   icode (Constructor a b c d e f)         = icode6 4 a b c d e f
   icode (Primitive   a b c)               = icode3 5 a b c
   value = vcase valu where
     valu [0, a]                            = valu1 Axiom       a
     valu [1, a, b, c, d, e, f]             = valu6 Function    a b c d e f
     valu [2, a, b, c, d, e, f, g, h, i, j] = valu10 Datatype    a b c d e f g h i j
-    valu [3, a, b, c, d, e, f, g, h]       = valu8 Record      a b c d e f g h
+    valu [3, a, b, c, d, e, f, g, h, i]    = valu9 Record      a b c d e f g h i
     valu [4, a, b, c, d, e, f]             = valu6 Constructor a b c d e f
     valu [5, a, b, c]                      = valu3 Primitive   a b c
-    valu _                                 = __IMPOSSIBLE__
+    valu _                                 = malformed
 
 instance EmbPrj FunctionInverse where
   icode NotInjective = icode0'
   icode (Inverse a)  = icode1' a
   value = vcase valu where valu []  = valu0 NotInjective
                            valu [a] = valu1 Inverse a
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
 
 instance EmbPrj TermHead where
   icode SortHead    = icode0 0
@@ -473,19 +594,19 @@ instance EmbPrj TermHead where
   value = vcase valu where valu [0]    = return SortHead
                            valu [1]    = return PiHead
                            valu [2, a] = valu1 ConHead a
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj Agda.Syntax.Common.IsAbstract where
   icode AbstractDef = icode0 0
   icode ConcreteDef = icode0 1
   value = vcase valu where valu [0] = valu0 AbstractDef
                            valu [1] = valu0 ConcreteDef
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
 
 instance EmbPrj I.Clause where
   icode (Clause a b c d e) = icode5' a b c d e
   value = vcase valu where valu [a, b, c, d, e] = valu5 Clause a b c d e
-                           valu _               = __IMPOSSIBLE__
+                           valu _               = malformed
 
 instance EmbPrj I.ClauseBody where
   icode (Body   a) = icode1 0 a
@@ -496,14 +617,14 @@ instance EmbPrj I.ClauseBody where
                            valu [1, a] = valu1 Bind   a
                            valu [2, a] = valu1 NoBind a
                            valu []     = valu0 NoBody
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj Delayed where
   icode Delayed    = icode0 0
   icode NotDelayed = icode0 1
   value = vcase valu where valu [0] = valu0 Delayed
                            valu [1] = valu0 NotDelayed
-                           valu _   = __IMPOSSIBLE__
+                           valu _   = malformed
 
 instance EmbPrj I.Pattern where
   icode (VarP a  ) = icode1 0 a
@@ -514,14 +635,14 @@ instance EmbPrj I.Pattern where
                            valu [1, a, b] = valu2 ConP a b
                            valu [2, a]    = valu1 LitP a
                            valu [3, a]    = valu1 DotP a
-                           valu _         = __IMPOSSIBLE__
+                           valu _         = malformed
 
 instance EmbPrj a => EmbPrj (Builtin a) where
   icode (Prim    a) = icode1 0 a
   icode (Builtin a) = icode1 1 a
   value = vcase valu where valu [0, a] = valu1 Prim    a
                            valu [1, a] = valu1 Builtin a
-                           valu _      = __IMPOSSIBLE__
+                           valu _      = malformed
 
 instance EmbPrj HP.NameKind where
   icode HP.Bound           = icode0 0
@@ -544,7 +665,7 @@ instance EmbPrj HP.NameKind where
     valu [6]     = valu0 HP.Postulate
     valu [7]     = valu0 HP.Primitive
     valu [8]     = valu0 HP.Record
-    valu _       = __IMPOSSIBLE__
+    valu _       = malformed
 
 instance EmbPrj HP.Aspect where
   icode HP.Comment       = icode0 0
@@ -563,7 +684,7 @@ instance EmbPrj HP.Aspect where
     valu [4]        = valu0 HP.Symbol
     valu [5]        = valu0 HP.PrimitiveType
     valu [6, mk, b] = valu2 HP.Name mk b
-    valu _          = __IMPOSSIBLE__
+    valu _          = malformed
 
 instance EmbPrj HP.OtherAspect where
   icode HP.Error              = icode0 0
@@ -578,26 +699,49 @@ instance EmbPrj HP.OtherAspect where
     valu [2] = valu0 HP.UnsolvedMeta
     valu [3] = valu0 HP.TerminationProblem
     valu [4] = valu0 HP.IncompletePattern
-    valu _   = __IMPOSSIBLE__
+    valu _   = malformed
 
 instance EmbPrj HP.MetaInfo where
   icode (HP.MetaInfo a b c d) = icode4' a b c d
 
   value = vcase valu where
     valu [a, b, c, d] = valu4 HP.MetaInfo a b c d
-    valu _            = __IMPOSSIBLE__
+    valu _            = malformed
 
-instance EmbPrj HP.HighlightingInfo where
-  icode (HP.HighlightingInfo a b) = icode2' a b
+instance EmbPrj Precedence where
+  icode TopCtx                 = icode0 0
+  icode FunctionSpaceDomainCtx = icode0 1
+  icode (LeftOperandCtx a)     = icode1 2 a
+  icode (RightOperandCtx a)    = icode1 3 a
+  icode FunctionCtx            = icode0 4
+  icode ArgumentCtx            = icode0 5
+  icode InsideOperandCtx       = icode0 6
+  icode WithFunCtx             = icode0 7
+  icode WithArgCtx             = icode0 8
+  icode DotPatternCtx          = icode0 9
+  value = vcase valu
+    where
+    valu [0]    = valu0 TopCtx
+    valu [1]    = valu0 FunctionSpaceDomainCtx
+    valu [2, a] = valu1 LeftOperandCtx a
+    valu [3, a] = valu1 RightOperandCtx a
+    valu [4]    = valu0 FunctionCtx
+    valu [5]    = valu0 ArgumentCtx
+    valu [6]    = valu0 InsideOperandCtx
+    valu [7]    = valu0 WithFunCtx
+    valu [8]    = valu0 WithArgCtx
+    valu [9]    = valu0 DotPatternCtx
+    valu _      = malformed
 
-  value = vcase valu where
-    valu [a, b] = valu2 HP.HighlightingInfo a b
-    valu _      = __IMPOSSIBLE__
+instance EmbPrj ScopeInfo where
+  icode (ScopeInfo a b c d) = icode4' a b c d
+  value = vcase valu where valu [a, b, c, d] = valu4 ScopeInfo a b c d
+                           valu _            = malformed
 
 instance EmbPrj Interface where
-  icode (Interface a b c d e f g) = icode7' a b c d e f g
-  value = vcase valu where valu [a, b, c, d, e, f, g] = valu7 Interface a b c d e f g
-                           valu _                     = __IMPOSSIBLE__
+  icode (Interface a b c d e f g h) = icode8' a b c d e f g h
+  value = vcase valu where valu [a, b, c, d, e, f, g, h] = valu8 Interface a b c d e f g h
+                           valu _                        = malformed
 
 
 
@@ -618,14 +762,16 @@ icodeN = icodeX nodeD nodeC
 
 vcase :: forall a . EmbPrj a => ([Int] -> R a) -> Int -> R a
 vcase valu ix = do
-    aTyp <- lift $ typeRepKey $ typeOf (undefined :: a)
-    memo <- asks nodeMemo
-    maybeU <- lift $ H.lookup memo (ix, aTyp)
+    memo <- gets nodeMemo
+    (aTyp, maybeU) <- liftIO $ do
+      aTyp   <- typeRepKey $ typeOf (undefined :: a)
+      maybeU <- H.lookup memo (ix, aTyp)
+      return (aTyp, maybeU)
     case maybeU of
-      Just (U u) -> maybe (__IMPOSSIBLE__) return (cast u)
+      Just (U u) -> maybe malformed return (cast u)
       Nothing    -> do
-          v <- valu . (! ix) =<< asks nodeE
-          lift $ H.insert memo (ix, aTyp) (U v)
+          v <- valu . (! ix) =<< gets nodeE
+          liftIO $ H.insert memo (ix, aTyp) (U v)
           return v
 
 icode0  tag                     = icodeN [tag]
@@ -664,19 +810,23 @@ valu8  z a b c d e f g h     = valu7 z a b c d e f g     `ap` value h
 valu9  z a b c d e f g h i   = valu8 z a b c d e f g h   `ap` value i
 valu10 z a b c d e f g h i j = valu9 z a b c d e f g h i `ap` value j
 
-test :: EmbPrj a => a -> IO a
-test x = decode =<< encode x
+-- | Creates an empty dictionary.
 
-emptyDict :: IO Dict
-emptyDict = liftM5 Dict
-            (H.new (==) hashNode)
-            (H.new (==) H.hashString)
-            (H.new (==) (H.hashInt . fromIntegral))
-            (H.new (==) (H.hashInt . floor))
-            (newIORef 0)
-            `ap` (newIORef 0)
-            `ap` (newIORef 0)
-            `ap` (newIORef 0)
+emptyDict :: SourceToModule
+             -- ^ Maps file names to the corresponding module names.
+             -- Must contain a mapping for every file name that is
+             -- later encountered.
+          -> IO Dict
+emptyDict fileMod = Dict
+  <$> H.new (==) hashNode
+  <*> H.new (==) H.hashString
+  <*> H.new (==) (H.hashInt . fromIntegral)
+  <*> H.new (==) (H.hashInt . floor)
+  <*> newIORef 0
+  <*> newIORef 0
+  <*> newIORef 0
+  <*> newIORef 0
+  <*> return fileMod
 
 hashNode :: [ Int ] -> Int32
 hashNode is = List.foldl' f golden is
